# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Supplies methods to convert ZX-graphs to tikz files.
These tikz files are designed to be easily readable by the program `Tikzit <https://tikzit.github.io>`_.
"""


import tempfile
import os
import subprocess
import shutil
import time
from fractions import Fraction
from typing import List, Dict, overload, Tuple, Union, Optional

from .utils import get_z_box_label, set_z_box_label, settings, EdgeType, VertexType, FloatInt
from .graph.base import BaseGraph, VT, ET
from .graph.graph import Graph
from .circuit import Circuit

TIKZ_BASE = """
\\begin{{tikzpicture}}
    \\begin{{pgfonlayer}}{{nodelayer}}
{vertices}
    \\end{{pgfonlayer}}
    \\begin{{pgfonlayer}}{{edgelayer}}
{edges}
    \\end{{pgfonlayer}}
\\end{{tikzpicture}}
"""

def _to_tikz(g: BaseGraph[VT,ET], draw_scalar:bool = False,
    xoffset:FloatInt=0, yoffset:FloatInt=0, idoffset:int=0
    ) -> Tuple[List[str],List[str]]:
    """Converts a ZX-graph ``g`` to a string representing a tikz diagram.
    The optional arguments are used by :func:`to_tikz_sequence`.
    """
    verts = []

    if draw_scalar:
        scalar = g.scalar.to_latex()
        x = -1 + xoffset + min([g.row(v) for v in g.vertices()],default=0)
        y = - yoffset - sum([g.qubit(v) for v in g.vertices()])/(g.num_vertices()+1)
        s = "        \\node [style=none] ({:d}) at ({:.2f}, {:.2f}) {{{:s}}};".format(idoffset,x,y,scalar)
        idoffset += 1
        verts.append(s)
    maxindex = idoffset

    for v in g.vertices():
        ty = g.type(v)
        if ty == VertexType.DUMMY:
            # Export dummy node with its text label and style=text
            text = g.vdata(v, 'text', '')
            style = settings.tikz_classes['dummy']
            x = g.row(v) + xoffset
            y = - g.qubit(v) - yoffset
            s = f"        \\node [style={style}] ({v+idoffset}) at ({{x:.2f}}, {{y:.2f}}) {{{{{text}}}}};".format(x=x, y=y)
            verts.append(s)
            maxindex = max([v+idoffset,maxindex])
            continue
        if ty == VertexType.Z_BOX:
            p = get_z_box_label(g,v)
        else:
            p = g.phase(v)
        if ty == VertexType.BOUNDARY:
            style = settings.tikz_classes['boundary']
        elif ty == VertexType.H_BOX:
            style = settings.tikz_classes['H']
        elif ty == VertexType.W_INPUT:
            style = settings.tikz_classes['W input']
        elif ty == VertexType.W_OUTPUT:
            style = settings.tikz_classes['W']
        elif ty == VertexType.Z_BOX:
            style = settings.tikz_classes['Z box']
        else:
            if p != 0:
                if ty==VertexType.Z: style = settings.tikz_classes['Z phase']
                else: style = settings.tikz_classes['X phase']
            else:
                if ty==VertexType.Z: style = settings.tikz_classes['Z']
                else: style = settings.tikz_classes['X']
        if ((ty == VertexType.H_BOX or ty == VertexType.Z_BOX) and p == 1) or\
            (ty != VertexType.H_BOX and p == 0):
            phase = ""
        elif type(p) == Fraction:
            ns = '' if p.numerator == 1 else str(p.numerator)
            dn = '' if p.denominator == 1 else str(p.denominator)
            if dn: phase = r"$\frac{%s\pi}{%s}$" % (ns, dn)
            else: phase = r"$%s\pi$" % ns
        else:
            phase = r"$%s$" % str(p)
        x = g.row(v) + xoffset
        y = - g.qubit(v) - yoffset
        s = "        \\node [style={}] ({:d}) at ({:.2f}, {:.2f}) {{{:s}}};".format(style,v+idoffset,x,y,phase)
        verts.append(s)
        maxindex = max([v+idoffset,maxindex])
    edges = []
    for e in g.edges():
        v,w = g.edge_st(e)
        et = g.edge_type(e)
        s = "        \\draw "
        if et == EdgeType.HADAMARD:
            if g.type(v) != VertexType.BOUNDARY and g.type(w) != VertexType.BOUNDARY:
                style = settings.tikz_classes['H-edge']
                if style: s += "[style={:s}] ".format(style)
            else:
                x = (g.row(v) + g.row(w))/2.0 +xoffset
                y = -(g.qubit(v)+g.qubit(w))/2.0 -yoffset
                t = "        \\node [style={:s}] ({:d}) at ({:.2f}, {:.2f}) {{}};".format(settings.tikz_classes['H'],maxindex+1, x,y)
                verts.append(t)
                maxindex += 1
        elif et == EdgeType.W_IO:
            style = settings.tikz_classes['W-io-edge']
            if style: s += "[style={:s}] ".format(style)
        else:
            style = settings.tikz_classes['edge']
            if style: s += "[style={:s}] ".format(style)
        s += "({:d}) to ({:d});".format(v+idoffset,w+idoffset)
        edges.append(s)

    return (verts, edges)

def to_tikz(g: BaseGraph[VT,ET], draw_scalar:bool=False) -> str:
    """Converts a ZX-graph ``g`` to a string representing a tikz diagram."""
    verts, edges = _to_tikz(g,draw_scalar)
    return TIKZ_BASE.format(vertices="\n".join(verts), edges="\n".join(edges))

def to_tikz_sequence(graphs:List[BaseGraph], draw_scalar:bool=False, maxwidth:FloatInt=10) -> str:
    """Given a list of ZX-graphs, outputs a single tikz diagram with the graphs presented in a grid.
    ``maxwidth`` is the maximum width of the diagram, before a graph is put on a new row in the tikz diagram."""
    xoffset = -maxwidth
    yoffset = -10
    idoffset = 0
    total_verts, total_edges = [],[]
    for g in graphs:
        max_index = max(g.vertices()) + 2*g.num_inputs() + 2
        verts, edges = _to_tikz(g,draw_scalar,xoffset,yoffset,idoffset)
        total_verts.extend(verts)
        total_edges.extend(edges)
        if xoffset + g.depth() + 2> maxwidth:
            xoffset = -maxwidth
            yoffset += g.qubit_count() + 2
        else:
            xoffset += g.depth() + 2
        idoffset += max_index

    return TIKZ_BASE.format(vertices="\n".join(total_verts), edges="\n".join(total_edges))



def tikzit(g: Union[BaseGraph[VT,ET],Circuit,str], draw_scalar:bool=False) -> None:
    """Opens Tikzit with the graph ``g`` opened as a tikz diagram. 
    For this to work, ``zx.settings.tikzit_location`` must be pointed towards the Tikzit executable.
    Even though this function is intended to be used with Tikzit, ``zx.tikz.tikzit_location``
    can point towards any executable that takes a tikz file as an input, such as a text processor."""

    if not settings.tikzit_location or shutil.which(settings.tikzit_location) is None:
        raise Exception("Please point towards the Tikzit executable"
                        " (or some other executable that accepts a text file as an argument)"
                        " with pyzx.settings.tikzit_location")

    if isinstance(g, Circuit):
        g = g.to_graph(zh=True)
    if isinstance(g, BaseGraph):
        tz = to_tikz(g,draw_scalar)
    else:
        tz = g
    with tempfile.TemporaryDirectory() as tmpdirname:
        fname = os.path.join(tmpdirname, "graph.tikz")
        with open(fname,'w') as f:
            f.write(tz)
        print("Opening Tikzit...")
        subprocess.check_call([settings.tikzit_location, fname])
        print("Done")


synonyms_boundary = ['none', 'empty', 'boundary']
synonyms_z = ['z dot', 'z spider', 'z', 'z phase dot',
              'white spider', 'white phase spider', 'white dot', 'white phase dot',
              'green dot', 'green node', 'green phase node']
synonyms_x = ['x dot', 'x spider', 'x', 'x phase dot',
              'grey spider', 'grey phase spider', 'grey dot', 'grey phase dot',
              'gray spider', 'gray phase spider', 'gray dot', 'gray phase dot',
              'red dot', 'red node', 'red phase node']
synonyms_hadamard = ['hadamard', 'h', 'small hadamard']
synonyms_w_input = ['w input']
synonyms_w_output = ['w output', 'w', 'w triangle']
synonyms_z_box = ['z box', 'zbox', 'zbox phase', 'green box', 'green box phase',
                  'green phase box', 'white box', 'white box phase', 'white phase box']
synonyms_dummy = ['text', 'label', 'dummy node', 'dummy spider', 'dummy phase spider']

synonyms_edge = ['empty', 'simple', 'none']
synonyms_hedge = ['hadamard edge']
synonyms_wedge = ['w edge', 'w io edge']

tikz_error_message = "Not a valid tikz picture. Please use Tikzit to generate correct output."
def tikz_to_graph(
    s: str,
    warn_overlap:bool= True,
    fuse_overlap:bool = True,
    ignore_nonzx:bool = False,
    backend:Optional[str]=None) -> BaseGraph:
    """Converts a tikz diagram into a pyzx Graph.
    The tikz diagram is assumed to be one generated by Tikzit,
    and hence should have a nodelayer and a edgelayer..

    Args:
        s: a string containing a well-defined Tikz diagram.
        warn_overlap: If True raises a Warning if two vertices have the exact same position.
        fuse_overlap: If True fuses two vertices that have the exact same position. Only has effect if fuse_overlap is False.
        ignore_nonzx: If True suppresses most errors about unknown vertex/edge types and labels.
        backend: Backend of the graph returned.

    Warning:
    	Vertices that might look connected in the output of the tikz are not necessarily connected
		at the level of tikz itself, and won't be treated as such in pyzx.
    """
    lines = [l.strip() for l in s.strip().splitlines() if l.strip() != '']
    if not lines[0].startswith(r'\begin{tikzpicture}'):
        raise ValueError(tikz_error_message)
    if lines[-1] != r'\end{tikzpicture}':
        raise ValueError(tikz_error_message)
    if lines[1] != r'\begin{pgfonlayer}{nodelayer}':
        raise ValueError(tikz_error_message)

    g = Graph(backend)
    index_dict: Dict[int,VT] = {} # type: ignore
    position_dict: Dict[str,List[int]] = {}
    for c,l in enumerate(lines[2:]):
        if l == r'\end{pgfonlayer}': break
        # l should look like
        # \node [style=stylename] (integer_id) at (x_float, y_float) {$phase$};
        if not l.startswith(r'\node'):
            raise ValueError(r"Node definition does not start with '\node': %s" % l)
        l = l[6:]
        i = l.find('[')
        if i == -1: raise ValueError("Node definition %s does not include style" % l)
        j = l.find(']',i)
        style = l[i+1:j].strip()
        style = style.split('=')[1]
        vs, other = l[j+1:].split("at")
        vid = int(vs.replace("(","").replace(")","").strip())
        pos, label = other.split("{",1)
        pos = pos.replace("(","").replace(")","").strip()
        x,y = [float(z) for z in pos.split(",")]
        label = label[:-2].replace('$','').replace(r'\ ','').replace('~','').strip()

        ty: VertexType
        if style.lower() in synonyms_boundary: ty = VertexType.BOUNDARY
        elif style.lower() in synonyms_z: ty = VertexType.Z
        elif style.lower() in synonyms_x: ty = VertexType.X
        elif style.lower() in synonyms_hadamard: ty = VertexType.H_BOX
        elif style.lower() in synonyms_w_input: ty = VertexType.W_INPUT
        elif style.lower() in synonyms_w_output: ty = VertexType.W_OUTPUT
        elif style.lower() in synonyms_z_box: ty = VertexType.Z_BOX
        elif style.lower() in synonyms_dummy: ty = VertexType.DUMMY
        else:
            if ignore_nonzx:
                ty = VertexType.BOUNDARY
            else:
                raise ValueError("Unknown vertex style '%s' in node definition %s" % (style, l))

        if pos in position_dict:
            if warn_overlap:
                raise Warning("Vertices %d and %s have same position" % (vid, str(position_dict[pos])))
            if fuse_overlap:
                v = index_dict[position_dict[pos][0]]
            else:
                v = g.add_vertex(ty,-y,x)
            position_dict[pos].append(vid)
        else:
            position_dict[pos] = [vid]
            v = g.add_vertex(ty,-y,x)
        index_dict[vid] = v

        if ty == VertexType.DUMMY:
            g.set_vdata(v, 'text', label)
            continue
        elif ty == VertexType.Z_BOX:
            set_phase = lambda v, p: set_z_box_label(g, v, p)
        else:
            set_phase = g.set_phase
        if label == '0':
            set_phase(v,0)
        elif label == r'\neg':
            set_phase(v,1)
        elif label:
            if label.find('pi') == -1 and ty != VertexType.Z_BOX:
                if not ignore_nonzx:
                    raise ValueError("Node definition %s has invalid phase label" % l)
            else:
                label = label.replace(r'\pi','').strip()
                if label == '' or label == '-' or label == '-1':
                    set_phase(v,1)
                elif label.find(r'\frac') != -1:
                    label = label.replace(r'\frac','').strip()
                    if label.find('}{') == -1:
                        n = 1
                        try:
                            m = int(label)
                        except:
                            raise ValueError("Node definition %s has invalid phase label" % l)
                    else:
                        num, denom = label.split('}{',1)
                        num = num.replace('{','').strip()
                        denom = denom.replace('}','').strip()
                        if num == '': n = 1
                        elif num == '-': n = -1
                        else:
                            try:
                                n = int(num)
                            except:
                                raise ValueError("Node definition %s has invalid phase label" % l)
                        try:
                            m = int(denom)
                        except:
                            raise ValueError("Node definition %s has invalid phase label" % l)
                    set_phase(v, Fraction(n,m))
                elif label.find('/') != -1:
                    num, denom = label.split('/',1)
                    if num == '': n = 1
                    elif num == '-': n = -1
                    else:
                        try:
                            n = int(num)
                        except:
                            raise ValueError("Node definition %s has invalid phase label" % l)
                    try:
                        m = int(denom)
                    except:
                        raise ValueError("Node definition %s has invalid phase label" % l)
                    set_phase(v, Fraction(n,m))
                else:
                    try:
                        if ty == VertexType.Z_BOX:
                            phase = complex(label)
                        else:
                            phase = int(label)
                    except:
                        raise ValueError("Node definition %s has invalid phase label '%s'" % (l,label))
                    set_phase(v, phase)

    # done parsing the vertices, now we parse the edges
    etab: Dict[ET, List[int]] = {} # type: ignore

    if lines[c+3] != r'\begin{pgfonlayer}{edgelayer}':
        raise ValueError(tikz_error_message)
    for c,l in enumerate(lines[c+4:]):
        if l == r'\end{pgfonlayer}': break
        if not l.startswith(r'\draw'):
            raise ValueError(r"Edge definition does not start with '\draw': %s" % l)
        l = l[6:]
        i = l.find('style')
        if i == -1:
            style = "empty"
            j = l.find(']')
        else:
            j1 = l.find(']')
            if j1 == -1:
                raise ValueError(r"Faulty edge definition %s" % l)
            j2 = l.find(',',i)
            if j2 != -1 and j2 < j1:
                style = l[i+5:j2].replace("=","").strip()
            else:
                style = l[i+5:j1].replace("=","").strip()
            j = j1
        src, tgt = l[j+1:].replace(".center","").split("to")
        src = src.replace("(","").replace(")","").strip()
        tgt = tgt.replace("(","").replace(")","").replace(";","").strip()

        e = index_dict[int(src)], index_dict[int(tgt)]

        if style.lower() in synonyms_edge:
            if e in etab:
                etab[e][0] += 1
            else:
                etab[e] = [1,0]
        elif style.lower() in synonyms_hedge:
            if e in etab:
                etab[e][1] += 1
            else:
                etab[e] = [0,1]
        elif style.lower() in synonyms_wedge:
            g.add_edge(e, EdgeType.W_IO)
        else:
            if ignore_nonzx:
                if e in etab:
                    etab[e][0] += 1
                else:
                    etab[e] = [1,0]
            else:
                raise ValueError("Unknown edge style '%s' in edge definition %s" % (style, l))

    g.add_edge_table(etab)
    return g



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict, Iterable, List, Optional, Union
from pyzx_param.circuit import Circuit, Gate, gate_types, CNOT
from pyzx_param.linalg import Z2, Mat2, MatLike

import numpy as np

class CNOT_tracker(Circuit):
    """
    A circuit-like object that keeps track of row and column operations
    applied during Gauss elimination via CNOT gates.
    """

    matrix: Mat2
    """The qubit parity matrix computed from the CNOT gates."""

    row_perm: np.ndarray
    """The row permutation of the qubit parity matrix."""

    col_perm: np.ndarray
    """The column permutation of the qubit parity matrix."""

    def __init__(self, n_qubits: int, **kwargs):
        """
        Creates and returns a circuit-like object.

        :param n_qubits: The number of qubits
        :param **kwargs: Additional arguments passed
        """
        super().__init__(n_qubits, **kwargs)
        #self.matrix = Mat2(np.identity(n_qubits, dtype=np.int32).tolist())
        self.matrix = Mat2.id(n_qubits)
        self.row_perm = np.arange(n_qubits)
        self.col_perm = np.arange(n_qubits)
        self.n_qubits = n_qubits

    def count_cnots(self) -> int:
        """Returns the number of CNOT gates in the tracker."""
        return len(
            [g for g in self.gates if hasattr(g, "name") and g.name in ["CNOT", "CZ"]]
        )

    def cnot_depth(self) -> int:
        """Returns the CNOT/CZ depth of the tracked circuit."""
        depth = 1
        previous_gates: List[int] = []
        for g in self.gates:
            if hasattr(g, "name") and g.name in ["CNOT", "CZ"]:
                if g.control in previous_gates or g.target in previous_gates:  # type: ignore # Overlapping gate
                    # Start a new CNOT layer
                    previous_gates = []
                    depth += 1
                else:
                    previous_gates += [g.control, g.target]  # type: ignore
        return depth

    def row_add(self, q0: int, q1: int):
        """Track a row addition operation on the matrix."""
        self.add_gate("CNOT", q0, q1)
        self.matrix.row_add(q0, q1)

    def add_gate(self, gate: Union[Gate,str], *args, **kwargs):
        """ Adds a gate to the circuit, if it is a CNOT gate it will track a row addition operation on the matrix. """
        if isinstance(gate, CNOT):
            self.row_add(gate.control, gate.target)
        else:
            super().add_gate(gate, *args, **kwargs)

    def col_add(self, q0: int, q1: int):
        """Track a column addition operation on the matrix."""
        self.prepend_gate("CNOT", q1, q0)
        self.matrix.col_add(q0, q1)

    @staticmethod
    def get_metric_names() -> List[str]:
        """Metric names for the CNOT tracker."""
        return ["n_cnots", "depth"]

    def gather_metrics(self) -> Dict[str, int]:
        """Gather metrics for the CNOT tracker."""
        metrics = {}
        metrics["n_cnots"] = self.count_cnots()
        metrics["depth"] = self.cnot_depth()
        return metrics

    def prepend_gate(self, gate: Union[Gate, str], *args, **kwargs):
        """Adds a gate to the circuit. ``gate`` can either be
        an instance of a :class:`Gate`, or it can be the name of a gate,
        in which case additional arguments should be given.

        Example::

            circuit.add_gate("CNOT", 1, 4) # adds a CNOT gate with control 1 and target 4
            circuit.add_gate("ZPhase", 2, phase=Fraction(3,4)) # Adds a ZPhase gate on qubit 2 with phase 3/4
        """
        if isinstance(gate, str):
            gate_class = gate_types[gate]
            gate = gate_class(*args, **kwargs)
        self.gates.insert(0, gate)

    def to_qasm(self, version: int = 2) -> str:
        """
        Outputs a QASM of the current circuit.

        :param version: The QASM version to be used, default 2
        :return: The initial wiring, the resulting wiring and the full QASM circuit
        """
        qasm = super().to_qasm(version)
        initial_perm = "// Initial wiring: " + str(self.row_perm)
        end_perm = "// Resulting wiring: " + str(self.col_perm)
        return "\n".join([initial_perm, end_perm, qasm])

    @staticmethod
    def from_circuit(circuit: Circuit) -> "CNOT_tracker":
        """
        Takes a circuit and outputs a CNOT_tracker of the circuit.

        :param circuit: A circuit to be converted
        :return: A CNOT_tracker of the circuit given
        """
        new_circuit = CNOT_tracker(circuit.qubits, name=circuit.name)
        new_circuit.gates = circuit.gates
        new_circuit.update_matrix()
        return new_circuit

    def update_matrix(self):
        """Rebuilds the parity matrix from the gates in the circuit."""
        self.matrix = Mat2.id(self.n_qubits)
        for gate in self.gates:
            if hasattr(gate, "name") and gate.name == "CNOT":
                self.matrix.row_add(gate.control, gate.target)
            else:
                print(
                    "Warning: CNOT tracker can only be used for circuits with only CNOT gates!"
                )

    @staticmethod
    def from_qasm_file(fname: str) -> "CNOT_tracker":
        """
        Takes the filename of a QASM and outputs a CNOT_tracker of it.

        :param fname: The filename of the QASM
        :return: A CNOT_tracker of the QASM given
        """
        circuit = Circuit.from_qasm_file(fname)
        return CNOT_tracker.from_circuit(circuit)

class Parity:
    """
    A set of qubits XORed together.
    """
    parity: List[bool]

    def __init__(self, par: Union[str, int, Iterable[Any]], n_qubits: Optional[int] = None):
        """
        Creates and returns a set of qubits XORed together.

        :param par: The input parity pattern
        :param n_qubits: An optional array of the number of qubits, default None
        """
        if isinstance(par, int):
            self.parity = [bool(par & (1 << i)) for i in reversed(range(par.bit_length()))]
        elif isinstance(par, str):
            self.parity = [p != "0" for p in par]
        else:
            self.parity = [bool(x) for x in par]
        if n_qubits is not None and len(self.parity) < n_qubits:
            self.parity += [False] * (n_qubits - len(self.parity))

    def count(self) -> int:
        """Returns the number of qubits interacting in the parity."""
        return sum(self.parity)

    def n_qubits(self) -> int:
        """Returns the total number of qubits."""
        return len(self.parity)
    
    def to_mat2_row(self) -> List[Z2]:
        """Returns an array of 1s and 0s based off the internal parity array."""
        return [1 if b else 0 for b in self.parity]

    def __str__(self):
        """Returns a string of 1s and 0s based off the internal parity array."""
        return "".join(["1" if p else "0" for p in self.parity])

    def __repr__(self):
        """Returns a string representation of the parity."""
        return "Parity(" + str(self) + ")"

    def __len__(self):
        """Returns the length of the parity."""
        return len(self.parity)

    def __iter__(self):
        """Makes the parity iterable."""
        return iter(self.parity)

    def __getitem__(self, i):
        """Returns the item at index i in the parity."""
        return self.parity[i]

    def __set_item__(self, i, v):
        """Sets the item at index i to the value given by v in the parity."""
        self.parity[i] = v

    def __eq__(self, other):
        """Returns true if other is equal to the parity."""
        if isinstance(other, Parity):
            return self.parity == other.parity
        if isinstance(other, str):
            return str(self) == other
        return False

    def __hash__(self):
        """Returns a hash of the parity."""
        return hash(str(self))



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import sys
import os
if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')

import io

from pyzx_param.scripts import main


class TestScripts(unittest.TestCase):

    def test_scripts_wrong_command(self):
        sys.stdout = io.StringIO()
        with self.assertRaises(SystemExit):
            main(['fakepath', 'bla', 'bla'])
        sys.stdout = sys.__stdout__

    def test_optimize_quipper_circuit_gives_quipper_circuit(self):
        sys.stdout = io.StringIO()
        main(['fakepath','opt','tests/test_circuit.circuit'])
        assert os.path.isfile('tests/test_circuit.quipper')
        os.remove('tests/test_circuit.quipper')
        sys.stdout = sys.__stdout__

    def test_optimize_quipper_to_qasm(self):
        sys.stdout = io.StringIO()
        main('fakepath opt -t qasm tests/test_circuit.circuit'.split())
        assert os.path.isfile('tests/test_circuit.qasm')
        os.remove('tests/test_circuit.qasm')
        sys.stdout = sys.__stdout__

    def test_optimize_all_options(self):
        sys.stdout = io.StringIO()
        main('fakepath opt -d tests/other_name.bla -t qc -g cliff tests/test_circuit.circuit'.split())
        assert os.path.isfile('tests/other_name.bla')
        os.remove('tests/other_name.bla')
        sys.stdout = sys.__stdout__

    def test_tikz_conversion(self):
        sys.stdout = io.StringIO()
        main('fakepath tikz tests/test_circuit.circuit tests/tikz_circuit.tikz'.split())
        assert os.path.isfile('tests/tikz_circuit.tikz')
        os.remove('tests/tikz_circuit.tikz')
        sys.stdout = sys.__stdout__
    

if __name__ == '__main__':
    print("Please only run this file as a collection of all tests, and with the working directory set to the parent directory.")


from . import Graph
from .graph.base import BaseGraph
from .symbolic import Poly
from .utils import VertexType, EdgeType, FractionLike

from fractions import Fraction
from typing import List, Tuple


def cluster_state(m: int, n: int, inputs: List[Tuple[int,int]]=[]) -> BaseGraph:
    """Build a cluster state m qubits tall and n qubits wide. Optionally, give a list of grid positions
    which serve as inputs."""
    g = Graph()
    vs = [[g.add_vertex(VertexType.Z, qubit=2*q,row=2*r) for r in range(n)] for q in range(m)]
    inp = []
    outp = []
    for q in range(m):
        for r in range(n):
            if q < m-1: g.add_edge((vs[q][r], vs[q+1][r]), edgetype=EdgeType.HADAMARD)
            if r < n-1: g.add_edge((vs[q][r], vs[q][r+1]), edgetype=EdgeType.HADAMARD)
            o = g.add_vertex(VertexType.BOUNDARY, qubit=2*q-0.8, row=2*r+0.8)
            g.add_edge((vs[q][r], o))
            outp.append(o)
            
            if (q,r) in inputs:
                i = g.add_vertex(VertexType.BOUNDARY, qubit=2*q+0.8, row=2*r-0.8)
                g.add_edge((vs[q][r], i))
                inp.append(i)

    g.set_inputs(tuple(inp))
    g.set_outputs(tuple(outp))
    return g

def measure(g:BaseGraph, pos:Tuple[int,int], t:VertexType=VertexType.Z, phase:FractionLike=0):
    """Measure the qubit at the given grid position, basis, and phase."""
    q = 2*pos[0]-0.8
    r = 2*pos[1]+0.8
    found = False
    if isinstance(phase, Poly):
        raise ValueError("Symbolic phases not supported")
    if not isinstance(phase, Fraction): phase = Fraction(phase)
    outputs = list(g.outputs())
    for v in g.vertices():
        if g.qubit(v) == q and g.row(v) == r:
            found = True
            if v in outputs:
                outputs.remove(v)
                g.set_type(v, t)
                g.set_phase(v, phase)
            else:
                raise ValueError("Already measured")
    g.set_outputs(tuple(outputs))
    if not found:
        raise ValueError("Couldn't find a qubit at that position")

def apply_pauli(g:BaseGraph, pos:Tuple[int,int], t:VertexType=VertexType.Z, phase:FractionLike=1):
    """Measure the qubit at the given grid position, basis, and phase."""

    if phase == 0:
        return
    elif phase != 1:
        raise ValueError("Phase should be 0 or 1")

    q = 2*pos[0]-0.8
    r = 2*pos[1]+0.8
    found = False
    outputs = g.outputs()
    verts = list(g.vertices())

    for v in verts:
        if g.qubit(v) == q and g.row(v) == r:
            found = True
            if v in outputs:
                ns = list(g.neighbors(v))
                if len(ns) == 1:
                    w = ns[0]
                    if t == g.type(w):
                        g.add_to_phase(w, 1)
                    else:
                        if t == VertexType.X:
                            g.remove_edge((v, w))
                            c = g.add_vertex(VertexType.X, qubit=q+0.4, row=r-0.4, phase=1)
                            g.add_edge((w, c))
                            g.add_edge((c, v))
                        else:
                            ns = [n for n in g.neighbors(w) if n != v]
                            if len(ns) == 1 and g.type(ns[0]) == VertexType.Z:
                                w1 = ns[0]
                                g.set_phase(w, -g.phase(w))
                                g.add_to_phase(w1, 1)
                            else:
                                raise ValueError("Can't perform a Z correction here")
                else:
                    raise ValueError("Expected a single neighbor for boundary vertex")
            else:
                raise ValueError("Already measured")

    if not found:
        raise ValueError("Couldn't find a qubit at that position")



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2024 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from .gflow import gflow
from .utils import EdgeType, VertexType, vertex_is_zx, phase_is_clifford
from .graph.base import BaseGraph, VT, ET

from typing import Any, Optional, Set, Dict, Tuple, Generic
import random

def multiply_paulis(p1: str, p2: str) -> str:
    if p1 == 'I': return p2
    elif p2 == 'I': return p1
    elif p1 == p2: return 'I'
    elif p1 != 'X' and p2 != 'X': return 'X'
    elif p1 != 'Y' and p2 != 'Y': return 'Y'
    elif p1 != 'Z' and p2 != 'Z': return 'Z'
    else: raise ValueError('Expected: I, X, Y, or Z')

def h_pauli(p: str) -> str:
    if p == 'I': return 'I'
    elif p == 'X': return 'Z'
    elif p == 'Y': return 'Y'
    else: return 'X'


class PauliWeb(Generic[VT, ET]):
    """A Pauli web
    
    This is a labelling of edges of a given graph by Paulis from the set {'X', 'Y', 'Z'}. In order
    to deal properly with Hadamard edges, we actually label all "half-edges", where a half-edge named
    (v, w) is the end of the edge closest to `v`, and (w, v) is the end closest to `w`.

    A Pauli web is "closed" if
    (i)  every spider has adjacent half-edges labelled by a stabiliser or anti-stabiliser of that spider
    (ii) for every half-edge (v,w) labelled by a Pauli P, the other half is labelled by P (for a normal edge)
         and labelled HPH (for a Hadamard edge).
    
    If these conditions are violated, the Pauli web is called "open", and a spider violating (i) or edge violating
    (ii) is called a boundary of the web.
    """
    def __init__(self, g: BaseGraph[VT,ET]):
        self.g = g
        self.es: Dict[Tuple[VT,VT], str] = dict()
    
    @staticmethod
    def random(g: BaseGraph[VT,ET], pX=0.1, pY=0.1, pZ=0.1):
        w = PauliWeb(g)
        for e in g.edges():
            s,t = g.edge_st(e)
            for _ in range(2):
                p = random.random()
                if p < pX:
                    w.add_half_edge((s,t), 'X')
                elif p < pX + pY:
                    w.add_half_edge((s,t), 'Y')
                elif p < pX + pY + pZ:
                    w.add_half_edge((s,t), 'Z')
                s,t = (t,s)
        return w
    
    def copy(self) -> PauliWeb:
        pw = PauliWeb(self.g)
        pw.es = self.es.copy()
        return pw
    
    def add_half_edge(self, v_pair: Tuple[VT, VT], pauli: str):
        s, t = v_pair
        p = multiply_paulis(self.es.get((s,t), 'I'), pauli)
        if p == 'I':
            self.es.pop((s,t),'')
        else:
            self.es[(s,t)] = p

    def add_edge(self, v_pair: Tuple[VT, VT], pauli: str):
        s, t = v_pair
        et = self.g.edge_type(self.g.edge(s, t))
        self.add_half_edge((s,t), pauli)
        self.add_half_edge((t,s), pauli if et == EdgeType.SIMPLE else h_pauli(pauli))
    
    def vertices(self):
        return set(v for (v,_) in self.es)

    def half_edges(self) -> Dict[Tuple[VT,VT],str]:
        return self.es
    
    def __repr__(self):
        return 'PauliWeb' + str(self.vertices())
    
    def __mul__(self, other: PauliWeb):
        pw = self.copy()
        for e,p in other.es.items():
            pw.add_half_edge(e, p)
        return pw
    
    def commutes_with(self, other: PauliWeb):
        comm = True
        for e,p1 in self.es.items():
            p2 = other.es.get(e, 'I')
            if p1 != 'I' and p2 != 'I' and p1 != p2:
                comm = not comm
        return comm
    
    def graph_with_errors(self) -> BaseGraph[VT,ET]:
        g = self.g.clone()
        edges = set((s,t) if s < t else (t,s) for s,t in self.es)
        for s,t in edges:
            p0 = self.es.get((s,t), 'I')
            p1 = self.es.get((t,s), 'I')

            e = g.edge(s, t)
            et = g.edge_type(e)
            g.remove_edge(e)

            spots = 1
            for p in [p0,p1]:
                if p == 'Y': spots += 2
                elif p == 'X' or p == 'Z': spots += 1

            q, r = (g.qubit(s), g.row(s))
            dq, dr = ((g.qubit(t) - q)/spots, (g.row(t) - r)/spots)

            v0 = s
            v1 = t
            if p0 == 'X':
                q += dq; r += dr
                v0 = g.add_vertex(VertexType.X, q, r, phase=1)
                g.add_edge((s, v0))
            elif p0 == 'Y':
                q += dq; r += dr
                v = g.add_vertex(VertexType.X, q, r, phase=1)
                q += dq; r += dr
                v0 = g.add_vertex(VertexType.Z, q, r, phase=1)
                g.add_edge((s, v))
                g.add_edge((v, v0))
            elif p0 == 'Z':
                q += dq; r += dr
                v0 = g.add_vertex(VertexType.Z, q, r, phase=1)
                g.add_edge((s, v0))

            if p1 == 'X':
                q += dq; r += dr
                v1 = g.add_vertex(VertexType.X, q, r, phase=1)
                g.add_edge((t, v1))
            elif p1 == 'Y':
                q += dq; r += dr
                v1 = g.add_vertex(VertexType.Z, q, r, phase=1)
                q += dq; r += dr
                v = g.add_vertex(VertexType.X, q, r, phase=1)
                g.add_edge((v1, v))
                g.add_edge((v, t))
            elif p1 == 'Z':
                q += dq; r += dr
                v1 = g.add_vertex(VertexType.Z, q, r, phase=1)
                g.add_edge((t, v1))
            
            g.add_edge((v0,v1), et)
        return g




# def transpose_corrections(c: Dict[VT, Set[VT]]) -> Dict[VT, Set[VT]]:
#     ct: Dict[VT, Set[VT]] = dict()
#     for k,s in c.items():
#         for v in s:
#             if v not in ct: ct[v] = set()
#             ct[v].add(k)
#     return ct

def compute_pauli_webs(g: BaseGraph[VT,ET], backwards:bool=True, debug:Optional[Dict[str,Any]]=None) -> Tuple[Dict[VT, int], Dict[VT, PauliWeb[VT,ET]], Dict[VT, PauliWeb[VT,ET]]]:
    g1 = g.clone()
    color_edge: Dict[VT, Tuple[VT,VT]] = dict()
    boundary: Dict[VT, VT] = dict()

    for e in g.edges():
        s, t = g.edge_st(e)
        st = g.type(s)
        tt = g.type(t)
        et = g.edge_type(e)
        if not (vertex_is_zx(st) and vertex_is_zx(tt)): continue

        if (st == tt and et == EdgeType.SIMPLE) or (st != tt and et == EdgeType.HADAMARD):
            v = g1.add_vertex(VertexType.Z if st == VertexType.X else VertexType.X)
            g1.remove_edge(e)
            g1.add_edge((s, v), EdgeType.SIMPLE)
            g1.add_edge((v, t), et)
            color_edge[v] = (s,t)
        
    for b in g1.inputs() + g1.outputs():
        e = g1.incident_edges(b)[0]
        v = next(iter(g1.neighbors(b)))
        vt = g1.type(v)
        et = g1.edge_type(e)
        g1.remove_edge(e)
        q, r = g1.qubit(b), g1.row(b)
        if ((vt == VertexType.Z and et == EdgeType.SIMPLE) or
            (vt == VertexType.X and et == EdgeType.HADAMARD)):
            v0 = g1.add_vertex(VertexType.X, qubit=q, row=r)
            v1 = g1.add_vertex(VertexType.Z, qubit=q, row=r)
        else:
            v0 = g1.add_vertex(VertexType.Z, qubit=q, row=r)
            v1 = g1.add_vertex(VertexType.X, qubit=q, row=r)
        g1.add_edge((v, v0), et)
        g1.add_edge((v0, v1), EdgeType.SIMPLE)
        g1.add_edge((v1, b), EdgeType.SIMPLE)
        color_edge[v0] = (b, v)
        boundary[v0] = b
        boundary[v1] = b

    if not debug is None:
        debug['g1'] = g1

    gf = gflow(g1, focus=True, reverse=backwards, pauli=True)
    if not gf:
        raise ValueError("Graph must have gFlow")
    order, corr = gf
    vset = g.vertex_set()
    order = { v: i for v,i in order.items() if v in vset and not phase_is_clifford(g.phase(v)) }

    zwebs: Dict[VT, PauliWeb[VT,ET]] = dict()
    xwebs: Dict[VT, PauliWeb[VT,ET]] = dict()
    for v,c in corr.items():
        pw = PauliWeb(g)
        for v1 in c:
            if v1 in vset:
                p = 'X' if g.type(v1) == VertexType.Z else 'Z'
                for v2 in g.neighbors(v1):
                    if g.type(v2) == VertexType.BOUNDARY or g1.connected(v1, v2):
                        pw.add_edge((v1, v2), p)
                    else:
                        pw.add_half_edge((v1, v2), p)
            elif v1 in color_edge:
                pw.add_edge(color_edge[v1], 'Z' if g1.type(v1) == VertexType.X else 'X')
        if v in boundary:
            ref = boundary[v]
        else:
            ref = v

        # pw.spread_to_boundary(inputs=True, outputs=False)
        if g1.type(v) == VertexType.Z: zwebs[ref] = pw
        elif g1.type(v) == VertexType.X: xwebs[ref] = pw

    return (order, zwebs, xwebs)



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['draw', 'arrange_scalar_diagram', 'draw_matplotlib', 'draw_d3',
           'draw_3d', 'matrix_to_latex', 'print_matrix', 'graphs_to_gif']

import os
import math
import cmath
import json
import string
import random
import importlib
from fractions import Fraction
from typing import Dict, List, Tuple, Optional, Union, Iterable, Any, TYPE_CHECKING
from typing_extensions import Literal
import numpy as np

from .pauliweb import PauliWeb

# matplotlib is lazy-imported on the first call to draw_matplotlib
plt: Any = None
path: Any = None
patches: Any = None
lines: Any = None


from .utils import settings, get_mode, phase_to_s, EdgeType, VertexType, FloatInt, get_z_box_label
from .graph.base import BaseGraph, VT, ET
from .circuit import Circuit

if get_mode() == "notebook":
    from IPython.display import display, HTML
elif get_mode() == "browser":
    from browser import document, html # type: ignore

if TYPE_CHECKING:
    from ipywidgets import Label

def draw(g: Union[BaseGraph[VT,ET], Circuit], labels: bool=False, **kwargs) -> Any:
    """Draws the given Circuit or Graph. 
    Depending on the value of ``pyzx.settings.drawing_backend``
    either uses matplotlib or d3 to draw."""

    # allow global setting to labels=False
    # TODO: probably better to make labels Optional[bool]
    labels = labels or settings.show_labels
    if get_mode() == "shell":
        return draw_matplotlib(g, labels, **kwargs)
    elif get_mode() == "browser":
        return draw_d3(g, labels, **kwargs)
    else: # in notebook
        if settings.drawing_backend == "d3":
            return draw_d3(g, labels, **kwargs)
        elif settings.drawing_backend == "matplotlib":
            return draw_matplotlib(g, labels, **kwargs)
        else:
            raise TypeError("Unsupported drawing backend '{}'".format(settings.drawing_backend))

def pack_circuit_nf(g: BaseGraph[VT,ET], nf:Literal['grg','gslc'] ='grg') -> None:
    x_index = 0
    ty = g.types()

    inputs = g.inputs()
    outputs = g.outputs()
    if nf == 'grg':
        for v in g.vertices():
            if v in inputs:
                g.set_row(v, 0)
            elif v in outputs:
                g.set_row(v, 4)
            elif ty[v] == VertexType.X:
                g.set_row(v, 2)
                g.set_qubit(v, x_index)
                x_index += 1
            elif ty[v] == VertexType.Z:
                for w in g.neighbors(v):
                    if w in inputs:
                        g.set_row(v,1)
                        g.set_qubit(v, g.qubit(w))
                        break
                    elif w in outputs:
                        g.set_row(v,3)
                        g.set_qubit(v, g.qubit(w))
                        break
    elif nf == 'gslc':
        for v in g.vertices():
            if v in inputs:
                g.set_row(v,0)
            elif v in outputs:
                g.set_row(v, 4)
            elif ty[v] == VertexType.Z:
                for w in g.neighbors(v):
                    if w in inputs:
                        g.set_row(v,1)
                        #g.set_vdata(v, 'q', g.get_vdata(w, 'q'))
                        break
                    elif w in outputs:
                        g.set_row(v,3)
                        #g.set_vdata(v, 'q', g.get_vdata(w, 'q'))
                        break
    else:
        raise ValueError("Unknown normal form: " + str(nf))

def arrange_scalar_diagram(g: BaseGraph[VT,ET]) -> None:
    g.normalize()
    rs = g.rows()
    qs = g.qubits()
    ty = g.types()
    gadgets: Dict[Tuple[VT,VT], FloatInt] = {}
    verts = []
    min_row = 1000000
    rows_used: Dict[FloatInt, List[VT]] = dict()
    for v in g.vertices():
        if len(list(g.neighbors(v))) == 1:
            w = list(g.neighbors(v))[0]
            gadgets[(v,w)] = 0
        elif all(g.vertex_degree(w) > 1 for w in g.neighbors(v)): # Not part of a phase gadget
            verts.append(v)
            #if rs[v] < min_row: min_row = rs[v]
            if rs[v] in rows_used: rows_used[rs[v]].append(v)
            else: rows_used[rs[v]] = [v]
    
    for i, r in enumerate(sorted(rows_used.keys())):
        for v in rows_used[r]:
            g.set_row(v,i)
            if qs[v] < 0: g.set_qubit(v,1)
            else: g.set_qubit(v,qs[v]+1)
    
    for v,w in gadgets.keys():
        score = sum(rs[n] for n in g.neighbors(w))/len(list(g.neighbors(w)))
        gadgets[(v,w)] = score
    
    l = list(gadgets.items())
    l = sorted(l, key=lambda x: x[1])
    for i in range(len(l)):
        v,w = l[i][0]
        g.set_row(v, i+0.5)
        g.set_row(w, i+0.5)
        g.set_qubit(v,-1)
        g.set_qubit(w,0)

def draw_matplotlib(
        g:      Union[BaseGraph[VT,ET], Circuit], 
        labels: bool                             =False, 
        figsize:Tuple[FloatInt,FloatInt]         =(8,2), 
        h_edge_draw: Literal['blue', 'box']      ='blue', 
        show_scalar: bool                        =False,
        rows: Optional[Tuple[FloatInt,FloatInt]] =None
        ) -> Any: # TODO: Returns a matplotlib figure

    # lazy import matplotlib
    global plt, path, patches, lines
    if plt is None:
        try:
            plt = importlib.import_module('matplotlib.pyplot')
            path = importlib.import_module('matplotlib.path')
            patches = importlib.import_module('matplotlib.patches')
            lines = importlib.import_module('matplotlib.lines')
        except ImportError:
            raise ImportError("This function requires matplotlib to be installed. "
                "If you are running in a Jupyter notebook, you can instead use `zx.draw_d3`.")

    if isinstance(g, Circuit):
        g = g.to_graph(zh=True)
    fig1, ax = plt.subplots(figsize=figsize)
    ax.set_frame_on(False)
    ax.set(xlim=(0, 1), ylim=(0, 1))
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)
    vs_on_row: Dict[FloatInt, int] = {} # count the vertices on each row
    for v in g.vertices():
        vs_on_row[g.row(v)] = vs_on_row.get(g.row(v), 0) + 1

    #Dict[VT,Tuple[FloatInt,FloatInt]]
    layout = {v:(g.row(v),-g.qubit(v)) for v in g.vertices()}

    if rows is not None:
        minrow,maxrow = rows
        vertices: Iterable[VT] = [v for v in g.vertices() if (minrow<=g.row(v) and g.row(v) <=maxrow)]
        edges: Iterable[ET] = [e for e in g.edges() if g.edge_s(e) in vertices and g.edge_t(e) in vertices]
    else:
        vertices = g.vertices()
        edges = g.edges()

    for e in edges:
        sp = layout[g.edge_s(e)]
        tp = layout[g.edge_t(e)]
        et = g.edge_type(e)
        n_row = vs_on_row.get(g.row(g.edge_s(e)), 0)

        dx = tp[0] - sp[0]
        dy = tp[1] - sp[1]
        bend_wire = (dx == 0) and h_edge_draw == 'blue' and n_row > 2
        if et == 2 and h_edge_draw == 'blue':
            ecol = '#0099ff'
        elif et == 3:
            ecol = 'gray'
        else:
            ecol = 'black'

        if bend_wire:
            bend = 0.25
            mid = (sp[0] + 0.5 * dx + bend * dy, sp[1] + 0.5 * dy - bend * dx)

            pth = path.Path([sp,mid,tp], [path.Path.MOVETO, path.Path.CURVE3, path.Path.LINETO])
            patch = patches.PathPatch(pth, edgecolor=ecol, linewidth=0.8, fill=False)
            ax.add_patch(patch)
        else:
            pos = 0.5 if dx == 0 or dy == 0 else 0.4
            mid = (sp[0] + pos*dx, sp[1] + pos*dy)
            ax.add_line(lines.Line2D([sp[0],tp[0]],[sp[1],tp[1]], color=ecol, linewidth=0.8, zorder=0))

        if h_edge_draw == 'box' and et == 2: #hadamard edge
            w = 0.2
            h = 0.15
            diag = math.sqrt(w*w+h*h)
            angle = math.atan2(dy,dx)
            angle2 = math.atan2(h,w)
            centre = (mid[0] - diag/2*math.cos(angle+angle2),
                      mid[1] - diag/2*math.sin(angle+angle2))
            ax.add_patch(patches.Rectangle(centre,w,h,angle=angle/math.pi*180,facecolor='yellow',edgecolor='black'))

        #plt.plot([sp[0],tp[0]],[sp[1],tp[1]], 'k', zorder=0, linewidth=0.8)

    for v in vertices:
        p = layout[v]
        t = g.type(v)
        a = g.phase(v)
        a_offset = 0.5
        phase_str = phase_to_s(a, t)

        if t == VertexType.Z:
            ax.add_patch(patches.Circle(p, 0.2, facecolor='#ccffcc', edgecolor='black', zorder=1))
        elif t == VertexType.X:
            ax.add_patch(patches.Circle(p, 0.2, facecolor='#ff8888', edgecolor='black', zorder=1))
        elif t == VertexType.H_BOX:
            ax.add_patch(patches.Rectangle((p[0]-0.1, p[1]-0.1), 0.2, 0.2, facecolor='yellow', edgecolor='black'))
            a_offset = 0.25
        elif t == VertexType.W_INPUT:
            ax.add_patch(patches.Circle(p, 0.05, facecolor='black', edgecolor='black', zorder=1))
        elif t == VertexType.W_OUTPUT:
            ax.add_patch(patches.Polygon([(p[0]-0.15, p[1]), (p[0]+0.15, p[1]+0.2), (p[0]+0.15, p[1]-0.2)], facecolor='black', edgecolor='black'))
        elif t == VertexType.Z_BOX:
            ax.add_patch(patches.Rectangle((p[0]-0.1, p[1]-0.1), 0.2, 0.2, facecolor='#ccffcc', edgecolor='black'))
            a_offset = 0.25
            phase_str = str(get_z_box_label(g, v))
        else:
            ax.add_patch(patches.Circle(p, 0.1, facecolor='black', edgecolor='black', zorder=1))

        if labels: plt.text(p[0]+0.25, p[1]+0.25, str(v), ha='center', color='gray', fontsize=5)
        if phase_str: plt.text(p[0], p[1]-a_offset, phase_str, ha='center', color='blue', fontsize=8)

    if show_scalar:
        x = min((g.row(v) for v in g.vertices()), default = 0)
        y = -sum((g.qubit(v) for v in g.vertices()))/(g.num_vertices()+1)
        ax.text(x-5,y,g.scalar.to_latex())

    ax.axis('equal')
    if get_mode() == "shell":
        plt.show()
    else:
        plt.close()
    return fig1
    

# Provides functions for displaying pyzx graphs in jupyter notebooks using d3

# make sure we get a fresh random seed
random_graphid = random.Random()

# def init_drawing() -> None:
#     if get_mode() not in ("notebook", "browser"): return
#
#     library_code = '<script type="text/javascript">\n'
#     for lib in ['d3.v5.min.inline.js']:
#         with open(os.path.join(settings.javascript_location, lib), 'r') as f:
#             library_code += f.read() + '\n'
#     library_code += '</script>'
#     display(HTML(library_code))

def auto_layout_vertex_locs(g:BaseGraph[VT, ET]): #Force-based graph drawing algorithm given by Eades(1984):
    c1 = 2 #Sample parameters that work decently well
    c2 = 1
    c3 = 1
    c4 = .1
    v_locs:Dict[VT, Tuple[float, float]] = dict()
    for v in g.vertices():
        v_locs[v]=(random.random()*math.sqrt(g.num_vertices()),  random.random()*math.sqrt(g.num_vertices()))
    for i in range(100): #100 iterations of force-based drawing
        forces:Dict[VT, Tuple[float, float]] = dict()
        for v in g.vertices():
            forces[v] = (0, 0)
            for v1 in g.vertices():
                if(v!=v1):
                    diff = (v_locs[v][0]-v_locs[v1][0], v_locs[v][1]-v_locs[v1][1])
                    d = math.sqrt(diff[0]*diff[0]+diff[1]*diff[1])
                    if g.connected(v1, v): #edge between vertices: apply rule c1*log(d/c2)
                        force_mag = -c1*math.log(d/c2) #negative force attracts
                    elif v != v1: #nonadjacent vertices: apply rule -c3/d^2
                        force_mag = c3/(d*d) #positive force repels
                    else: #free body in question, applies no force on itself
                        raise ValueError("Vertices ended up at same point")
                    v_force = (diff[0]*force_mag*c4/d, diff[1]*force_mag*c4/d)
                    forces[v] = (forces[v][0]+v_force[0], forces[v][1]+v_force[1])
        for v in g.vertices(): #leave y value constant if input or output
            v_locs[v]=(v_locs[v][0]+forces[v][0], v_locs[v][1]+forces[v][1])
    max_x = max(v[0] for v in v_locs.values())
    min_x = min(v[0] for v in v_locs.values())
    max_y = max(v[1] for v in v_locs.values())
    min_y = min(v[1] for v in v_locs.values())
    v_locs = {k:(v[0]-min_x, v[1]-min_y) for k, v in v_locs.items()} #translate to origin
    return v_locs, max_x-min_x, max_y-min_y

def graph_json(g: BaseGraph[VT, ET],
               coords: Dict[VT, Tuple[Any, Any, Any]],
               vdata: Optional[List[str]]=None,
               pauli_web: Optional[PauliWeb[VT,ET]]=None) -> str:

    nodes = [{'name': str(v),
              'x': float(coords[v][0]),
              'y': float(coords[v][1]),
              'z': float(coords[v][2]),
              't': g.type(v),
              'phase': g.get_phase_str(v) if g.type(v) != VertexType.Z_BOX else str(get_z_box_label(g, v)),
              'ground': g.is_ground(v),
              'vdata': [(key, g.vdata(v, key))
                  for key in vdata or [] if g.vdata(v, key, None) is not None],
              }
             for v in g.vertices()]

    # keep track of the number of parallel edges seen for a source/target pair
    counts: Dict[Tuple[str,str], int] = dict()
    links = []
    for e in g.edges():
        s = str(g.edge_s(e))
        t = str(g.edge_t(e))
        i = counts.get((s,t), 0)
        links.append({'source': s,
                      'target': t,
                      't': g.edge_type(e),
                      'index': i })
        counts[(s,t)] = i + 1
    for link in links:
        s,t = (str(link['source']), str(link['target']))
        link['num_parallel'] = counts[(s,t)]

    if pauli_web:
        pw_edges = [{ 'source': vs[0], 'target': vs[1], 't': t } for vs,t in pauli_web.half_edges().items() ]
    else:
        pw_edges = []

    return json.dumps({'nodes': nodes, 'links': links, 'pauli_web': pw_edges})


def draw_d3(
    g: Union[BaseGraph[VT,ET], Circuit],
    labels:bool=False, 
    scale:Optional[FloatInt]=None, 
    auto_hbox:Optional[bool]=None,
    show_scalar:bool=False,
    vdata: Optional[List[str]]=None,
    pauli_web: Optional[PauliWeb[VT,ET]]=None,
    auto_layout: bool = False
    ) -> Any:
    """If auto_layout is checked, will automatically space vertices of graph
    with no regard to qubit/row."""
    if get_mode() not in ("notebook", "browser"): 
        raise Exception("This method only works when loaded in a webpage or Jupyter notebook")

    if auto_hbox is None:
        auto_hbox = settings.drawing_auto_hbox

    if isinstance(g, Circuit):
        g = g.to_graph(zh=True)

    # tracking global sequence can cause clashes if you restart the kernel without clearing ouput, so
    # use an 8-digit random alphanum instead.
    graph_id = ''.join(random_graphid.choice(string.ascii_letters + string.digits) for _ in range(8))

    if(auto_layout):
        v_dict, w, h = auto_layout_vertex_locs(g)
        if scale is None:
            scale = 800 / w
            if scale > 50: scale = 50
            if scale < 20: scale = 20
        
        w = (w+2) * scale
        h = (h+3) * scale
    else:
        minrow = min([g.row(v) for v in g.vertices()], default=0)
        maxrow = max([g.row(v) for v in g.vertices()], default=0)
        minqub = min([g.qubit(v) for v in g.vertices()], default=0)
        maxqub = max([g.qubit(v) for v in g.vertices()], default=0)

        if scale is None:
            scale = 800 / (maxrow-minrow + 2)
            if scale > 50: scale = 50
            if scale < 20: scale = 20

        w = (maxrow-minrow + 2) * scale
        h = (maxqub-minqub + 3) * scale

    node_size = 0.2 * scale
    if node_size < 2: node_size = 2

    coords = {v:((v_dict[v][0]+1)*scale if auto_layout else (g.row(v)-minrow + 1) * scale,
               (v_dict[v][1]+2)*scale if auto_layout else (g.qubit(v)-minqub + 2) * scale,
               0) for v in g.vertices()}

    graphj = graph_json(g, coords, vdata=vdata, pauli_web=pauli_web)

    with open(os.path.join(settings.javascript_location, 'zx_viewer.inline.js'), 'r') as f:
        library_code = f.read() + '\n'

    text = """<div style="overflow:auto; background-color: white" id="graph-output-{id}"></div>
<script type="module">
var d3;
if (d3 == null) {{ d3 = await import("https://esm.sh/d3@5"); }}
var _settings_colors = JSON.parse('{colors}');
{library_code}
showGraph('#graph-output-{id}',
  JSON.parse('{graph}'), {width}, {height}, {scale},
  {node_size}, {hbox}, {labels}, '{scalar_str}');
</script>""".format(colors=json.dumps(settings.colors),
                    library_code=library_code,
                    id = graph_id,
                    graph = graphj, 
                    width=w, height=h, scale=scale, node_size=node_size,
                    hbox = 'true' if auto_hbox else 'false',
                    labels='true' if labels else 'false',
                    scalar_str=g.scalar.to_unicode() if show_scalar else '')
    if get_mode() == "notebook":
        display(HTML(text))
    else:
        d = html.DIV(style={"overflow": "auto", "background-color": "white"}, id="graph-output-{}".format(graph_id))
        source = """
        require(['zx_viewer'], function(zx_viewer) {{
            zx_viewer.showGraph('#graph-output-{0}',
            JSON.parse('{1}'), {2}, {3}, {4}, false, false);
        }});
        """.format(graph_id, graphj, str(w), str(h), str(node_size))
        s = html.SCRIPT(source, type="text/javascript")
        return d,s


def draw_3d(
    g: Union[BaseGraph[VT,ET], Circuit],
    labels: bool=False, 
    pauli_web: Optional[PauliWeb[VT,ET]]=None
    ) -> Any:

    if isinstance(g, Circuit):
        g = g.to_graph(zh=False)

    graph_id = ''.join(random_graphid.choice(string.ascii_letters + string.digits) for _ in range(8))

    minx = 0.0
    maxx = 0.0
    miny = 0.0
    maxy = 0.0
    minz = 0.0
    maxz = 0.0
    for v in g.vertices():
        minx = min(g.row(v), minx)
        maxx = max(g.row(v), maxx)
        miny = min(g.qubit(v), miny)
        maxy = max(g.qubit(v), maxy)
        minz = min(g.vdata(v, 'z', default=0), minz)
        maxz = max(g.vdata(v, 'z', default=0), maxz)
        
    coords = {v:(g.row(v) - (minx+maxx)/2,
               g.qubit(v) - (miny+maxy)/2,
               g.vdata(v, 'z', default=0) - (minz+maxz)/2)
               for v in g.vertices()}
    graphj = graph_json(g, coords, pauli_web=pauli_web)

    with open(os.path.join(settings.javascript_location, 'zx_viewer3d.js'), 'r') as f:
        library_code = f.read() + '\n'

    html_str = f"""
    <div style="overflow:auto; background-color: white" id="graph-output-3d-{graph_id}"></div>
    <script type="importmap">
    {json.dumps(settings.javascript_importmap)}
    </script>
    <script type="module">
    let _settings_colors = JSON.parse('{json.dumps(settings.colors)}');
    {library_code}
    showGraph3D('graph-output-3d-{graph_id}', JSON.parse('{graphj}'), {1000}, {500}, {'true' if labels else 'false'});
    </script>
    """
    # print(html_str)
    display(HTML(html_str))

# The dictionaries below are needed to
# pretty-print complex numbers in pretty_complex() and matrix_to_latex()

special_vals = {
    1: "1",
    -1: "-1",
    math.sqrt(2): r"\sqrt{2}",
    math.sqrt(3): r"\sqrt{3}",
    -math.sqrt(2): r"-\sqrt{2}",
    -math.sqrt(3): r"-\sqrt{3}",
    math.sqrt(1/2): r"\frac{1}{\sqrt{2}}",
    -math.sqrt(1/2): r"-\frac{1}{\sqrt{2}}",
    # math.sqrt(2) - 1:                 r"(\sqrt{2}-1)",
    # 1-math.sqrt(2):                 r"(1-\sqrt{2})",
    # 1+math.sqrt(2):                 r"(\sqrt{2}+1)",
    # 1+math.sqrt(1/2):                 r"(1+\frac{1}{\sqrt{2}})",
    # 1-math.sqrt(1/2):                 r"(1-\frac{1}{\sqrt{2}})",
    0.5*math.sqrt(1+math.sqrt(1/2)):r"\frac12\sqrt{1+\frac{1}{\sqrt{2}}}"
}

simple_vals = {
    1: "1",
    1/2: r"\frac12",
    1/3: r"\frac13",
    1/4: r"\frac14",
    2: "2",
    3: "3"
}
sqrt_vals = {
    math.sqrt(2): r"\sqrt{2}",
    math.sqrt(3): r"\sqrt{3}",
    math.sqrt(3)/2: r"\frac12\sqrt{3}",
    math.sqrt(1/2): r"\frac{1}{\sqrt{2}}",
    2*math.sqrt(2): r"2\sqrt{2}",
    math.sqrt(1/2)/2: r"\frac{1}{2\sqrt{2}}",
    math.sqrt(1/2)/4: r"\frac{1}{2\sqrt{4}}"
}

for v,s in simple_vals.items():
    for w,t in sqrt_vals.items():
        special_vals[v+w] = f"\\left({s}+{t}\\right)"
        special_vals[v-w] = f"\\left({s}-{t}\\right)"
        special_vals[-v+w] = f"\\left({t}-{s}\\right)"
        special_vals[-v-w] = f"-\\left({t}+{s}\\right)"

def strip_brackets(s:str) -> str:
    if s.startswith("(") and s.endswith(")"):
        return s[1:-1]
    if s.startswith("\\left(") and s.endswith("\\right)"):
        return s[6:-7]
    return s

def pretty_complex(z: complex) -> str:
    """Pretty print a complex number. Suitable for including in a
    Jupyter widgets Label()."""
    if abs(z) < 0.0000001:
        return "0"
    for v, s in special_vals.items():
        if abs(z-v) < 0.0001:
            return s
    out = ""
    r, arg = cmath.polar(z)
    farg = Fraction(arg/math.pi).limit_denominator(64) # this is now a fraction between -1 and 1
    if abs(farg*math.pi - arg) > 0.00001: # Polar decomp is not a good choice
        f = int(math.log10(1/abs(z)))
        if abs(f) > 1:
            z *= 10**f
        a,b = z.real, z.imag
        real_part, imag_part = False, False
        for v,s in special_vals.items():
            if abs(a-v) < 0.0001:
                out += s
                real_part = True
                break
        else:
            if abs(a) > 0.001:
                if abs(round(a)-a) < 0.0001:
                    out += str(round(a))
                elif abs(round(a*math.sqrt(2))-a*math.sqrt(2)) < 0.0001:
                    v = a/math.sqrt(2)
                    out += f"{round(v):d}\\sqrt{{2}}"
                else:
                    out += f"{a:.2f}".rstrip("0").rstrip(".")
                real_part = True

        if abs(b+1) < 0.0001:
            out += "-i"
            imag_part = True
        else:
            for v,s in special_vals.items():
                if abs(b-v) < 0.0001:
                    if b > 0:
                        out += "+"
                    out += s + "i"
                    imag_part = True
                    break
            else:
                if abs(b) > 0.001:
                    if b > 0.0:
                        if abs(a) > 0.001: out += "+"
                    if b < 0.0:
                        out += "-"
                        b = -b
                    if abs(b-1) < 0.001:
                        out += ""
                    else:
                        if abs(round(b)-b) < 0.0001:
                            out += str(round(b))
                        elif abs(round(b*math.sqrt(2))-b*math.sqrt(2)) < 0.0001:
                            v = b/math.sqrt(2)
                            out += f"{round(v):d}\\sqrt{{2}}"
                        else:
                            out += f"{b:.2f}".rstrip("0").rstrip(".")
                    out += "i"
                    imag_part = True
        if abs(f) > 1:
            out = f"({out})10^{{{-f}}}"
        elif real_part and imag_part:
            out = f"({out})"
        return out
    arg = farg # type: ignore
    if abs(r-1) > 0.00001: # not close to 1
        for v,s in special_vals.items():
            if abs(r-v) < 0.00001:
                out += s
                break
        else:
            v = math.log(r,2)
            vfrac = Fraction(v).limit_denominator(64)
            if abs(v-vfrac) < 0.00001:
                if vfrac.denominator == 1:
                    if vfrac == 1:
                        out += "2"
                    else:
                        out += r"2^{%d}" % vfrac.numerator
                else:
                    vfrac *= 2
                    if vfrac.denominator == 1:
                        if vfrac == 1:
                            out += r"\sqrt{2}"
                        else:
                            out += r"\sqrt{2}^{%d}" % vfrac.numerator
                    else:
                        out += r"\sqrt{2}^{\frac{%d}{%d}}" % (vfrac.numerator,vfrac.denominator)
            else:
                f = int(math.log10(1/abs(r)))
                if abs(f) > 1:
                    r *= 10**f
                if abs(round(r)-r) < 0.0001:
                    out += str(round(r))
                elif abs(round(r*math.sqrt(2))-r*math.sqrt(2)) < 0.0001:
                    v = r/math.sqrt(2)
                    out += f"{round(v):d}\\sqrt{{2}}"
                else:
                    out += f"{r:.2f}".rstrip("0").rstrip(".")
                if abs(f) > 1:
                    out += f"\\cdot 10^{{{-f}}}"
    
    minus = ""
    if arg < 0:
        minus = "-"
        arg += 1
    if arg == 1:
        minus = "-"
    elif arg == Fraction(1,2):
        out += "i"
    elif arg != 0:
        out += r"e^{i\frac{%d}{%d}\pi}" % (arg.numerator, arg.denominator) # type: ignore
    out = minus + out
    
    return out

def matrix_to_latex(m: np.ndarray) -> str:
    """Converts a matrix into latex code.
    Useful for pretty printing the matrix of a Circuit/Graph.

    Example:
        # Run this in a Jupyter notebook
        from ipywidgets import Label
        c = zx.Circuit(3)
        display(Label(matrix_to_latex(c.to_matrix())))
    """
    out = "\\begin{equation}\n"

    epsilon = 10**-14
    best_val = None
    denom: Optional[int] = None
    for v in m.flat:
        if abs(v) > epsilon:
            if best_val is None: 
                best_val = v
                denom = Fraction(cmath.phase(v)/math.pi).limit_denominator(512).denominator
            else:
                p = Fraction(cmath.phase(v)/math.pi).limit_denominator(512)
                if denom is None or p.denominator < denom:
                    best_val = v
                    denom = p.denominator
    if best_val is None:
        # matrix is zero
        out += "\\begin{pmatrix}\n"
        out += " \\\\ \n".join([" & ".join("0" for a in row) for row in m])
        out += "\n\\end{pmatrix}\n\\end{equation}"
        return out
    
    v = best_val
    m = m/v
    
    s = pretty_complex(v)
    if s == "1": s = ""
    if s == "-1": s= "-"
    out += s + "\n\\begin{pmatrix}\n"
    out += " \\\\ \n".join([" & ".join(strip_brackets(pretty_complex(a)) for a in row) for row in m])
    out += "\n\\end{pmatrix}\n\\end{equation}"
    return out


def print_matrix(m: Union[np.ndarray,BaseGraph,Circuit]) -> None:
    """Display a Label() Jupyter widget with a pretty LaTeX representation of the given matrix.
    Instead of a matrix, can also give a Circuit or Graph.
    """
    if get_mode() != "notebook":
        raise TypeError("Unsupported mode for print_matrix: '{}'".format(get_mode()))

    from ipywidgets import Label
    if isinstance(m, BaseGraph) or isinstance(m, Circuit):
        m = m.to_matrix()

    display(Label(matrix_to_latex(m)))


def graphs_to_gif(graphs: List[BaseGraph], filename: str, frame_duration: float=0.5):
    """Given a list of graphs, outputs an animated gif showing them in sequence.

    Args:
        graphs: The list of Graph instances that should be made into a gif.
        filename: the full filename of the output gif.
        frame_duration: how long (in seconds) each frame should last.

    Warning:
        This function requires imagio to be installed (pip install imageio).

    """
    import tempfile
    from pathlib import Path
    try:
        import imageio # type: ignore
    except ImportError:
        raise Exception("This function requires imageio to be installed (try: pip install imageio).")

    with tempfile.TemporaryDirectory() as tmpdirname:
        #print(tz)
        for i,g in enumerate(graphs):
            fig = draw_matplotlib(g)
            fname = os.path.join(tmpdirname, "graph{:03d}.png".format(i))
            fig.savefig(fname)
        image_path = Path(tmpdirname)
        images = list(image_path.glob('*.png'))
        image_list = []
        for file_name in images:
            image_list.append(imageio.imread(file_name))
        durations = [frame_duration]*len(image_list)
        durations[-1] = 5*frame_duration
        imageio.mimwrite(filename, image_list, duration=durations)
        return os.path.abspath(filename)



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the Scalar class used to represent a global scalar in a Graph."""

import math
import cmath
import copy
from fractions import Fraction
from typing import List, Optional, Set, Dict
import json

from ..utils import FloatInt, FractionLike

__all__ = ['Scalar']

def cexp(val) -> complex:
    return cmath.exp(1j*math.pi*val)

unicode_superscript = {
    '0': '⁰',
    '1': '¹',
    '2': '²',
    '3': '³',
    '4': '⁴',
    '5': '⁵',
    '6': '⁶',
    '7': '⁷',
    '8': '⁸',
    '9': '⁹'
}

unicode_fractions = {
    Fraction(1,4): '¼',
    Fraction(1,2): '½',
    Fraction(3,4): '¾',
}

class DyadicNumber:
    k: int
    a: int
    b: int
    c: int
    d: int

    def __init__(self, k: int = 0, a: int = 0, b: int = 0, c: int = 0, d: int = 0):

        while a % 2 == 0 and b % 2 == 0 and c % 2 == 0 and d % 2 == 0:
            a //= 2
            b //= 2
            c //= 2
            d //= 2
            k -= 1

        self.k = k
        self.a = a
        self.b = b
        self.c = c
        self.d = d

    def to_complex(self) -> complex:
        return (
            self.a
            + self.b * cmath.exp(1j * math.pi / 4)
            + self.c * 1j
            + self.d * cmath.exp(-1j * math.pi / 4)
        ) / (2**self.k)

    def __mul__(self, other: "DyadicNumber") -> "DyadicNumber":
        return DyadicNumber(
            self.k + other.k,
            self.a * other.a + self.b * other.d - self.c * other.c + self.d * other.b,
            self.a * other.b + self.b * other.a + self.c * other.d + self.d * other.c,
            self.a * other.c + self.b * other.b + self.c * other.a - self.d * other.d,
            self.a * other.d - self.b * other.c - self.c * other.b + self.d * other.a,
        )

    @staticmethod
    def sqrt2() -> "DyadicNumber":
        return DyadicNumber(0, 0, 1, 0, 1)

    @staticmethod
    def one() -> "DyadicNumber":
        return DyadicNumber(0, 1, 0, 0, 0)

    def conjugate(self) -> "DyadicNumber":
        return DyadicNumber(self.k, self.a, self.d, -self.c, self.b)

    def copy(self) -> "DyadicNumber":
        return DyadicNumber(self.k, self.a, self.b, self.c, self.d)

class SpiderPair:
    def __init__(self, alpha, beta, paramsA, paramsB):
        self.alpha:   int      = alpha   # phase n of n*pi/4 (i.e. = 0,1,2,3,4,5,6,7) #TODO: Can change this to fraction if we want to support any alpha
        self.beta:    int      = beta    # phase n of n*pi/4 (i.e. = 0,1,2,3,4,5,6,7)
        self.paramsA: Set[str] = paramsA # the set of XOR'd variables added to alpha
        self.paramsB: Set[str] = paramsB # the set of XOR'd variables added to beta
        # gamma   = (alpha + beta) % 2
        # paramsC = XOR(paramsA, paramsB)

class Scalar(object):
    """Represents a global scalar for a Graph instance."""
    def __init__(self) -> None:
        self.power2: int = 0 # Stores power of square root of two
        self.phase: Fraction = Fraction(0) # Stores complex phase of the number
        #self.phasevars: Dict[str,int] = dict() # Stores the phase variables. e.g. variable p: stores px/2, where x=0,1,2,3;  Dict[p,x]
        self.phasevars_pi: Set[str] = set() # Stores the basic phase variable terms with pi coefficients (c=pi)
        self.phasevars_pi_pair: List[List[Set[str]]] = [] # Stores the AND-pair phase variable term pairs with pi coefficients (c=pi)
        self.phasevars_halfpi: Dict[int, List[Set[str]]] = dict() # Stores the phase variable terms with +-pi/2 coeffs: c[terms[vars]], where c=1 or 3 (pi/2 or 3pi/2); These arise from lcomp's
        self.phasepairs: List[SpiderPair] = [] # Stores list of spider-pairs
        self.phasenodes: List[FractionLike] = [] # Stores list of legless spiders, by their phases.
        self.phasenodevars: List[Set[str]] = [] # Stores the added parameters of the legless spider phases
        self.floatfactor: DyadicNumber = DyadicNumber.one()
        self.approximate_floatfactor: complex = 1.0 + 0.0j
        self.is_unknown: bool = False # Whether this represents an unknown scalar value
        self.is_zero: bool = False

    def __repr__(self) -> str:
        return "Scalar({})".format(str(self))

    def __str__(self) -> str:
        if self.is_unknown:
            return "UNKNOWN"
        s = "{0.real:.2f}{0.imag:+.2f}i = ".format(self.to_number())
        if self.floatfactor.to_complex() != 1.0:
            s += "{0.real:.2f}{0.imag:+.2f}i".format(self.floatfactor.to_complex())
        if self.phase:
            s += "exp({}ipi)".format(str(self.phase))
        s += "sqrt(2)^{:d}".format(self.power2)
        for node in self.phasenodes:
            s += "(1+exp({}ipi))".format(str(node))
        return s

    def print_attrs(self):
        s = f"phasenode: {self.phasenodes}, {self.phasenodevars}"
        s += f"\nphasevars_pi_pair: {self.phasevars_pi_pair}"
        s += f"\nphasepairs: {[(pp.alpha, pp.beta, pp.paramsA, pp.paramsB) for pp in self.phasepairs]}"
        s += f"\nphasevars_halfpi: {self.phasevars_halfpi}"
        s += f"\nphase: {self.phase}"
        s += f"\npower2: {self.power2}"
        s += f"\nfloatfactor: {self.floatfactor.to_complex()}"
        s += f"\napproximate_floatfactor: {self.approximate_floatfactor}"
        s += f"\nis_zero: {self.is_zero}"
        print(s)
        return s

    
    def print_expr(self) -> str:
        scalar = self
        scalar_str = ""

        def format_phase_str(alpha, params):
            a_str = str(alpha) if alpha != 0 else ""
            for vars in params:
                a_str += f"+{vars}"
            if len(a_str) > 0 and a_str[0] == "+":
                a_str = a_str[1:]
            return a_str

        for const, vars in zip(scalar.phasenodes, scalar.phasenodevars):
            scalar_str += f"(1 + exp(iπ {format_phase_str(const, vars)}))"

        for pp in scalar.phasepairs:
            a_str = format_phase_str(pp.alpha / 4, pp.paramsA)
            b_str = format_phase_str(pp.beta / 4, pp.paramsB)

            scalar_str += (
                f"(1 + exp(iπ {a_str}) + exp(iπ {b_str}) - exp(iπ ({a_str} + {b_str})))"
            )

        for c in [1, 3]:
            if c not in scalar.phasevars_halfpi:
                continue
            for vars in scalar.phasevars_halfpi[c]:
                a_str = " + ".join(vars)
                scalar_str += f"exp(iπ {a_str} * {c}/4)"

        for pp in scalar.phasevars_pi_pair:
            if len(pp[0]) == 0 or len(pp[1]) == 0:
                continue
            a_str = " + ".join(pp[0])
            b_str = " + ".join(pp[1])

            scalar_str += f"exp(iπ {a_str} * {b_str})"

        if scalar.power2 % 2 == 0:
            if scalar.power2 > 0:
                scalar_str += f" * {2 ** (scalar.power2 // 2)}"
            elif scalar.power2 < 0:
                scalar_str += f" / {2 ** ((-scalar.power2) // 2)}"
        else:
            scalar_str += f" * sqrt(2) ** {scalar.power2}"

        print(scalar_str)
        return scalar_str

    def evaluate_scalar(self, vals: dict[str, Fraction]) -> complex:
        scalar = self
        number = 1

        vals["1"] = Fraction(1)

        # phase nodes
        for const, vars in zip(scalar.phasenodes, scalar.phasenodevars):
            number *= 1 + cexp(const + sum(vals[var] for var in vars))

        # phase pairs
        for pp in scalar.phasepairs:
            psi = pp.alpha / 4 + sum(vals[var] for var in pp.paramsA)
            phi = pp.beta / 4 + sum(vals[var] for var in pp.paramsB)
            number *= 1 + cexp(psi) + cexp(phi) - cexp(psi + phi)

        # half-pi
        for c in [1, 3]:
            if c not in scalar.phasevars_halfpi:
                continue
            for vars in scalar.phasevars_halfpi[c]:
                number *= cexp((sum(vals[var] for var in vars) % 2) * c / 2)

        # pi-pair
        for pp in scalar.phasevars_pi_pair:
            psi = sum(vals[var] for var in pp[0])
            phi = sum(vals[var] for var in pp[1])
            number *= cexp(psi * phi)

        if scalar.is_zero:
            return 0

        number *= cexp(scalar.phase)

        number *= math.sqrt(2) ** scalar.power2
        number *= scalar.floatfactor.to_complex()
        number *= scalar.approximate_floatfactor

        return number

    def __complex__(self) -> complex:
        return self.to_number()

    def polar_str(self) -> str:
        """Returns a human-readable string of the scalar in polar format"""
        r,th = cmath.polar(self.to_number())
        s = "{:.3}".format(r)
        if th != 0:
            s += " * exp(i pi * {:.3})".format(th/math.pi)
        return s

    def copy(self, conjugate: bool = False) -> 'Scalar':
        s = Scalar()
        s.power2 = self.power2
        s.phase = self.phase if not conjugate else -self.phase
        s.phasevars_pi = copy.copy(self.phasevars_pi)
        
        #TEMP:
        s.phasevars_pi_pair = []
        for i in self.phasevars_pi_pair:
            psA = copy.copy(i[0])
            psB = copy.copy(i[1])
            s.phasevars_pi_pair.append([psA,psB])
        
        #TEMP:
        s.phasevars_halfpi: Dict[List[Set[str]]] = dict()
        if 1 in self.phasevars_halfpi:
            for i in self.phasevars_halfpi[1]:
                if 1 not in s.phasevars_halfpi: s.phasevars_halfpi[1] = list()
                s.phasevars_halfpi[1].append(i)
        if 3 in self.phasevars_halfpi:
            for i in self.phasevars_halfpi[3]:
                if 3 not in s.phasevars_halfpi: s.phasevars_halfpi[3] = list()
                s.phasevars_halfpi[3].append(i)
        
        s.phasepairs = copy.copy(self.phasepairs)
        s.phasenodes = copy.copy(self.phasenodes) if not conjugate else [-p for p in self.phasenodes]
        s.phasenodevars = copy.copy(self.phasenodevars)
        s.floatfactor = self.floatfactor.copy() if not conjugate else self.floatfactor.conjugate()
        s.approximate_floatfactor = self.approximate_floatfactor if not conjugate else self.approximate_floatfactor.conjugate()
        s.is_unknown = self.is_unknown
        s.is_zero = self.is_zero
        return s

    def conjugate(self) -> 'Scalar':
        """Returns a new Scalar equal to the complex conjugate"""
        return self.copy(conjugate=True)

    def to_number(self) -> complex:
        if self.is_zero: return 0
        val = cexp(self.phase)
        for node in self.phasenodes: # Node should be a Fraction
            val *= 1+cexp(node)
        for sp in self.phasepairs:
            # alpha, beta are stored as integer multiples of pi/4
            a = Fraction(sp.alpha, 4)
            b = Fraction(sp.beta, 4)
            val *= 1 + cexp(a) + cexp(b) - cexp(a + b)
        val *= math.sqrt(2)**self.power2
        return val*self.floatfactor.to_complex() * self.approximate_floatfactor

    def to_latex(self) -> str:
        """Converts the Scalar into a string that is compatible with LaTeX."""
        if self.is_zero: return "0"
        elif self.is_unknown: return "Unknown"
        f = self.floatfactor.to_complex()
        for node in self.phasenodes:
            f *= 1+cexp(node)
        if self.phase == 1:
            f *= -1

        s = "$"
        if abs(f+1) < 0.001: #f \approx -1
            s += "-"
        elif abs(f-1) > 0.0001: #f \neq 1
            s += str(self.floatfactor.to_complex())
        if self.power2 != 0:
            s += r"\sqrt{{2}}^{{{:d}}}".format(self.power2)
        if self.phase not in (0,1):
            s += r"\exp(i~\frac{{{:d}\pi}}{{{:d}}})".format(self.phase.numerator,self.phase.denominator)
        s += "$"
        if s == "$$": return ""
        return s

    def to_unicode(self) -> str:
        """Returns a representation of the scalar that uses unicode
        to represent pi's and sqrt's."""
        if self.is_zero: return "0"
        elif self.is_unknown: return "Unknown"
        f = self.floatfactor.to_complex()
        for node in self.phasenodes:
            f *= 1+cexp(node)
        phase = Fraction(self.phase)
        if self.phase >= 1:
            f *= -1
            phase -= 1

        if abs(f+1) > 0.001 and abs(f-1) > 0.001:
            return str(f)

        s = ""
        if abs(f+1) < 0.001: #f \approx -1
            s += "-"
        if self.power2 != 0:
            s += r"√2"
            if self.power2 < 0:
                s += "⁻"
            val = str(abs(self.power2))
            s += "".join([unicode_superscript[i] for i in val])
        if phase != 0:
            s += "exp(i"
            if phase in unicode_fractions:
                s += unicode_fractions[phase] + "π)"
            else:
                s += "{:d}/{:d}π)".format(phase.numerator,phase.denominator)
        return s

    def to_dict(self) -> Dict[str, any]:
        d = {"power2": self.power2, "phase": str(self.phase)}
        if abs(self.floatfactor.to_complex() - 1) > 0.00001:
            d["floatfactor"] =  self.floatfactor.to_complex()
        if self.phasenodes:
            d["phasenodes"] = [str(p) for p in self.phasenodes]
        if self.is_zero:
            d["is_zero"] = self.is_zero
        if self.is_unknown:
            d["is_unknown"] = self.is_unknown,
        return d

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, s):
        if isinstance(s, str):
            d = json.loads(s)
        else:
            d = s
        # print('scalar from json', repr(d))
        scalar = Scalar()
        scalar.phase = Fraction(d["phase"]) # TODO support parameters
        scalar.power2 = int(d["power2"])
        if "floatfactor" in d:
            scalar.floatfactor = complex(d["floatfactor"])
        if "phasenodes" in d:
            scalar.phasenodes = [Fraction(p) for p in d["phasenodes"]]
        if "is_zero" in d:
            scalar.is_zero = bool(d["is_zero"])
        if "is_unknown" in d:
            scalar.is_unknown = bool(d["is_unknown"])
        return scalar

    def set_unknown(self) -> None:
        self.is_unknown = True
        self.phasenodes = []

    def add_power(self, n) -> None:
        """Adds a factor of sqrt(2)^n to the scalar."""
        self.power2 += n
    def add_phase(self, phase: FractionLike) -> None:
        """Multiplies the scalar by a complex phase."""
        self.phase = (self.phase + phase) % 2
    def add_phase_vars_halfpi(self, ps:Set[str], c:int) -> None: # These terms arise from lcomp's
        """Adds a term of XOR'd phase variables to the multiplier, for a +-pi/2 coefficient"""
        if (len(ps)>0): # Don't bother adding empty sets (i.e. those with no parameters)
            if c not in self.phasevars_halfpi: self.phasevars_halfpi[c] = list()
            self.phasevars_halfpi[c].append(ps)
    def add_phase_vars_pi(self, psA:Set[str]) -> None:
        """Adds XOR'd phase variables to the multiplier, for a pi coefficient"""
        self.phasevars_pi = self.phasevars_pi.symmetric_difference(psA)
    def add_phase_vars_pi_pair(self, psA:Set[str], psB:Set[str]) -> None:
        """Adds a term of XOR'd phase variable set pairs to the multiplier, for a pi coefficient"""
        if len(psA) == 0 or len(psB) == 0:
            return
        self.phasevars_pi_pair.append([psA, psB])
    def add_phase_pair(self, alpha: FractionLike, beta: FractionLike, paramsA: Set[str], paramsB: Set[str]) -> None:
        """Add a new spider-pair scalar term"""
        a = int(alpha*4) # Convert via alpha=a*pi/4
        b = int(beta*4)  # Convert via  beta=b*pi/4
        assert a == alpha * 4, f"alpha: {alpha}, a: {a}"
        assert b == beta * 4, f"beta: {beta}, b: {b}"
        sp = SpiderPair(a, b, paramsA, paramsB)
        self.phasepairs.append(sp)
        
    def add_node(self, node: FractionLike, node_params: Optional[Set[str]] = None) -> None:
        """A solitary spider with a phase ``node`` is converted into the
        scalar 1+e^(i*pi*node)."""
        if node_params is None:
            node_params = set()
        if (node == 0 and len(node_params) == 0):
            self.power2 += 2
        else:
            assert node.denominator in [1,2,4]
            self.phasenodes.append(node)
            self.phasenodevars.append(node_params) # XOR
        if (node == 1 and len(node_params) == 0): self.is_zero = True
    def add_float(self,f: DyadicNumber) -> None:
        # Expect DyadicNumber throughout; keep it closed under multiplication.
        if not isinstance(f, DyadicNumber):
            raise TypeError("floatfactor must remain a DyadicNumber")
        self.floatfactor *= f

    def mult_with_scalar(self, other: 'Scalar') -> None:
        """Multiplies two instances of Scalar together."""
        self.power2 += other.power2
        self.phase = (self.phase +other.phase)%2
        self.phasevars_pi       = self.phasevars_pi.symmetric_difference(other.phasevars_pi)
        for i in other.phasevars_pi_pair: self.phasevars_pi_pair.append(i) # TODO: Make this a deep copy for safety; currently we don't modify s.phasevars_pi_pair anywhere, so shallow copy is fine
        for c in other.phasevars_halfpi: 
            if (c in self.phasevars_halfpi):
                self.phasevars_halfpi[c].append( set().union(other.phasevars_halfpi[c]) )
            else: 
                self.phasevars_halfpi[c] = []
                self.phasevars_halfpi[c] = self.phasevars_halfpi[c] + copy.copy(other.phasevars_halfpi[c])
        self.phasepairs.extend(other.phasepairs)
        self.phasenodes.extend(other.phasenodes)
        self.phasenodevars.extend(other.phasenodevars)
        if not isinstance(other.floatfactor, DyadicNumber):
            raise TypeError("floatfactor must remain a DyadicNumber")
        self.floatfactor *= other.floatfactor
        self.approximate_floatfactor *= other.approximate_floatfactor
        if other.is_zero: self.is_zero = True
        if other.is_unknown: self.is_unknown = True

    def add_spider_pair(self, p1: FractionLike,p2: FractionLike, params1:Set[str], params2:Set[str]) -> None:
        """Add the scalar corresponding to a connected pair of spiders (p1)-H-(p2)."""
        # These if statements look quite arbitrary, but they are just calculations of the scalar
        # of a pair of connected single wire spiders of opposite colors.
        # We make special cases for Clifford phases and pi/4 phases.
        self.add_power(-1)
        self.add_phase_pair(p1,p2,params1,params2)



#!/usr/bin/python
# type: ignore
import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "readme.md").read_text()

setup(
    name="pyzx-param",
    author="Quantomatic",
    author_email="john@vdwetering.name",
    version="0.9.1",
    url="https://github.com/rafaelha/pyzx.git",
    description="Library for quantum circuit rewriting and optimisation using the ZX-calculus",
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache2",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 4 - Beta",
    ],
    packages=[
        "pyzx_param",
        "pyzx_param.circuit",
        "pyzx_param.graph",
        "pyzx_param.rewrite_rules",
        "pyzx_param.routing",
        "pyzx_param.local_search",
        "pyzx_param.scripts",
    ],
    python_requires='>=3.9',
    install_requires=["typing_extensions>=3.7.4",
                      "numpy>=1.14",
                      "pyperclip>=1.8.1",
                      "tqdm>=4.56.0",
                      "lark>=1.2.2"],
    include_package_data=True,
)



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import random
import sys
from types import ModuleType
from typing import Optional

from pyzx_param import VertexType, EdgeType

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
from pyzx_param.graph import Graph
from pyzx_param.circuit import Circuit
from pyzx_param.circuit.qasmparser import qasm
from fractions import Fraction
from pyzx_param.generate import cliffordT
from pyzx_param.simplify import *
from pyzx_param.simplify import supplementarity_simp, to_clifford_normal_form_graph, copy_simp
from pyzx_param import compare_tensors
from pyzx_param.generate import cliffordT

np: Optional[ModuleType]
try:
    import numpy as np
    from pyzx_param.tensor import tensorfy, compare_tensors
except ImportError:
    np = None

SEED = 1337


@unittest.skipUnless(np, "numpy needs to be installed for this to run")
class TestSimplify(unittest.TestCase):

    def setUp(self):
        random.seed(SEED)
        self.circuits = []
        self.circuits.append(cliffordT(3,20,0.3))
        self.circuits.append(cliffordT(3,10,0.1))
        # self.circuits.append(cliffordT(4,30,0.3))
        # self.circuits.append(cliffordT(5,50,0.08))
        # self.circuits.append(cliffordT(4,80,0.1))

    def func_test(self, func, prepare=None):
        for i,c in enumerate(self.circuits):
            with self.subTest(i=i, func=func.__name__):
                if prepare:
                    for f in prepare: f(c,quiet=True)
                t = tensorfy(c)
                func(c, quiet=True)
                t2 = tensorfy(c)
                self.assertTrue(compare_tensors(t,t2))
                del t, t2

    def test_spider_simp(self):
        self.func_test(spider_simp)

    def test_id_simp(self):
        self.func_test(id_simp)

    def test_to_gh(self):
        self.func_test(to_gh)

    def test_pivot_simp(self):
        self.func_test(pivot_simp,prepare=[spider_simp,to_gh,spider_simp])

    def test_lcomp_simp(self):
        self.func_test(lcomp_simp,prepare=[spider_simp,to_gh,spider_simp])

    def test_clifford_simp(self):
        self.func_test(clifford_simp)

    def test_supplementarity_simp(self):
        g = Graph()
        v = g.add_vertex(1,0,0,phase=Fraction(1,4))
        w = g.add_vertex(1,1,0,phase=Fraction(7,4))
        g.add_edge((v,w),2)
        vs = []
        for i in range(3):
            h = g.add_vertex(1,i,2,Fraction(1))
            vs.append(h)
            g.add_edges([(v,h),(w,h)],2)
        t = g.to_tensor()
        i = supplementarity_simp(g,quiet=True)
        self.assertEqual(i,1)
        self.assertTrue(compare_tensors(t,g.to_tensor()))

    def test_teleport_reduce(self):
        """Tests whether teleport_reduce preserves semantics on a set of circuits that have been broken before."""
        for i,s in enumerate([qasm_1,qasm_2,qasm_3,qasm_4]):
            with self.subTest(i=i):
                c = qasm(s)
                g = c.to_graph()
                c2 = Circuit.from_graph(teleport_reduce(g))
                self.assertTrue(c.verify_equality(c2))

    def test_to_graph_like_introduce_boundary_vertices(self):
        c = qasm(qasm_5)
        g = c.to_graph()
        to_graph_like(g)
        self.assertTrue(compare_tensors(c,g))

    def test_full_reduce_with_h_box(self):
        """Test that calls to :func:`full_reduce` with a graph containing H-boxes raises an error.
        This is a common mistake made by users (e.g., see issues #161 and #200).
        """
        g = Graph()
        v0 = g.add_vertex(VertexType.BOUNDARY, 0, 0)
        v1 = g.add_vertex(VertexType.H_BOX, 0, 1)
        v2 = g.add_vertex(VertexType.BOUNDARY, 0, 2)
        g.add_edge((v0, v1))
        g.add_edge((v1, v2))

        with self.assertRaises(ValueError) as context:
            full_reduce(g)
        self.assertTrue("Input graph is not a ZX-diagram" in str(context.exception))

    def test_full_reduce_scalar(self):
        """Test that checks whether a scalar is correctly removed from a graph using full_reduce.
        """

        from pyzx_param import Graph, full_reduce 
        g = Graph()
        g.add_vertex(ty=1, phase=0.5)
        g.add_vertex(ty=1, phase=1)
        g.add_edge((0, 1))

        full_reduce(g)

        g1 = Graph()
        g1.add_vertex(ty=1, phase=1)

        full_reduce(g1)
        
        self.assertTrue(g.num_vertices() == 0)
        self.assertTrue(g1.num_vertices() == 0)


    def test_to_clifford_normal_form_graph(self):
        for _ in range(10):
            g = cliffordT(4, 20, p_t=0)
            g0 = g.copy()
            to_clifford_normal_form_graph(g)
            self.assertTrue(compare_tensors(g0, g, preserve_scalar=True))
    
    def test_copy_simp(self):
        g = Graph() 

        v0 = g.add_vertex(VertexType.Z, 0, 0)
        v1 = g.add_vertex(VertexType.X, 0, 1)
        g.add_edge((v0, v1))
        v2 = g.add_vertex(VertexType.BOUNDARY, 0, 3)
        g.add_edge((v1, v2))
        v3 = g.add_vertex(VertexType.BOUNDARY, 1, 3)
        g.add_edge((v1, v3))

        g1 = g.copy()
        to_gh(g1)
        copy_simp(g1)

        g1.auto_detect_io()
        g.auto_detect_io()

        self.assertFalse(g.num_vertices() != g1.num_vertices())
        self.assertTrue(g1.num_vertices() == 4)
        self.assertTrue(compare_tensors(g1.to_tensor(),g.to_tensor()))

    
    def test_copy_simp_full_reduce(self):
        g = Graph() 

        v0 = g.add_vertex(VertexType.Z, 0, 0)
        v1 = g.add_vertex(VertexType.X, 0, 1)
        g.add_edge((v0, v1))
        v2 = g.add_vertex(VertexType.BOUNDARY, 0, 3)
        g.add_edge((v1, v2))
        v3 = g.add_vertex(VertexType.BOUNDARY, 1, 3)
        g.add_edge((v1, v3))

        g1 = g.copy()
        to_gh(g)
        copy_simp(g)

        full_reduce(g1)
        g.auto_detect_io()
        g1.auto_detect_io()

        self.assertTrue(g.num_vertices() == g1.num_vertices())
        self.assertTrue(compare_tensors(g1.to_tensor(),g.to_tensor()))

    def test_copy_simp_with_interacting_matches(self):
        g = Graph()
        v0 = g.add_vertex(VertexType.Z, 0, 0)
        v1 = g.add_vertex(VertexType.Z, 0, 1, Fraction(1, 1))
        v2 = g.add_vertex(VertexType.Z, 0, 2, Fraction(1, 1))
        v3 = g.add_vertex(VertexType.Z, 0, 3)

        g.add_edges([(v0, v1), (v1, v2), (v2, v3)], edgetype=EdgeType.HADAMARD)

        t1 = g.to_tensor()
        copy_simp(g)
        t2 = g.to_tensor()
        self.assertTrue(np.allclose(t1, t2))



qasm_1 = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[3];
x q[2];
ccx q[0], q[2], q[1];
x q[2];
ccx q[2], q[1], q[0];
cx q[2], q[0];
x q[2];
ccx q[0], q[2], q[1];
ccx q[0], q[2], q[1];
"""

qasm_2 = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[3];
ccx q[0], q[1], q[2];
ccx q[0], q[1], q[2];
ccx q[1], q[0], q[2];
x q[0];
ccx q[1], q[0], q[2];
ccx q[2], q[1], q[0];
ccx q[2], q[1], q[0];
ccx q[1], q[0], q[2];
"""

qasm_3 = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[3];
ccx q[1], q[2], q[0];
x q[0];
x q[1];
ccx q[1], q[0], q[2];
ccx q[1], q[0], q[2];
ccx q[1], q[0], q[2];
ccx q[0], q[1], q[2];
ccx q[0], q[2], q[1];
"""

qasm_4 = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[3];
ccx q[1], q[2], q[0];
x q[1];
ccx q[2], q[1], q[0];
ccx q[2], q[1], q[0];
ccx q[0], q[1], q[2];
ccx q[0], q[2], q[1];
ccx q[2], q[0], q[1];
cx q[1], q[2];
"""

qasm_5 = """OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
h q[0];
h q[1];
cz q[0],q[1];
"""


if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the ZX-diagram simplification strategies of PyZX.
Each strategy is based on applying some combination of the rewrite rules in the rules_ module.
The main procedures of interest are :func:`clifford_simp` for simple reductions,
:func:`full_reduce` for the full rewriting power of PyZX, and :func:`teleport_reduce` to
use the power of :func:`full_reduce` while not changing the structure of the graph.
"""

__all__ = ['bialg_simp','spider_simp', 'id_simp', 'phase_free_simp', 'pivot_simp',
        'pivot_gadget_simp', 'pivot_boundary_simp', 'gadget_simp',
        'lcomp_simp', 'clifford_simp', 'tcount', 'to_gh', 'to_rg',
        'full_reduce', 'teleport_reduce', 'reduce_scalar', 'supplementarity_simp',
        'to_clifford_normal_form_graph', 'to_graph_like', 'is_graph_like']

from ast import Mult
from functools import reduce
from optparse import Option
from typing import List, Callable, Optional, Union, Generic, Tuple, Dict, Iterator, cast

from .utils import EdgeType, VertexType, phase_is_clifford, toggle_edge, vertex_is_zx, toggle_vertex
from .rewrite_rules.rules import *
from .graph.base import BaseGraph, VT, ET
from .graph.multigraph import Multigraph
from .circuit import Circuit

class Stats(object):
    def __init__(self) -> None:
        self.num_rewrites: Dict[str,int] = {}
    def count_rewrites(self, rule: str, n: int) -> None:
        if rule in self.num_rewrites:
            self.num_rewrites[rule] += n
        else:
            self.num_rewrites[rule] = n
    def __str__(self) -> str:
        s = "REWRITES\n"
        nt = 0
        for r,n in self.num_rewrites.items():
            nt += n
            s += "%s %s\n" % (str(n).rjust(6),r)
        s += "%s TOTAL" % str(nt).rjust(6)
        return s


def simp(
    g: BaseGraph[VT,ET],
    name: str,
    match: Callable[..., List[MatchObject]],
    rewrite: Callable[[BaseGraph[VT,ET],List[MatchObject]],RewriteOutputType[VT,ET]],
    auto_simplify_parallel_edges: bool = False,
    matchf:Optional[Union[Callable[[ET],bool], Callable[[VT],bool]]]=None,
    quiet:bool=True,
    stats:Optional[Stats]=None) -> int:
    """Helper method for constructing simplification strategies based on the rules present in rules_.
    It uses the ``match`` function to find matches, and then rewrites ``g`` using ``rewrite``.
    If ``matchf`` is supplied, only the vertices or edges for which matchf() returns True are considered for matches.

    Example:
        ``simp(g, 'spider_simp', rules.match_spider_parallel, rules.spider)``

    Args:
        g: The graph that needs to be simplified.
        str name: The name to display if ``quiet`` is set to False.
        match: One of the ``match_*`` functions of rules_.
        rewrite: One of the rewrite functions of rules_.
        auto_simplify_parallel_edges: whether to automatically combine parallel edges between vertices if the graph is a Multigraph
        matchf: An optional filtering function on candidate vertices or edges, which
           is passed as the second argument to the match function.
        quiet: Suppress output on numbers of matches found during simplification.

    Returns:
        Number of iterations of ``rewrite`` that had to be applied before no more matches were found."""

    auto_simp_value = g.get_auto_simplify()
    if auto_simplify_parallel_edges:
        g.set_auto_simplify(True)
    i = 0
    new_matches = True
    while new_matches:
        new_matches = False
        if matchf is not None:
            m = match(g, matchf)
        else:
            m = match(g)
        if len(m) > 0:
            i += 1
            if i == 1 and not quiet: print("{}: ".format(name),end='')
            if not quiet: print(len(m), end='')
            #print(len(m), end='', flush=True) #flush only supported on Python >3.3
            etab, rem_verts, rem_edges, check_isolated_vertices = rewrite(g, m)
            g.add_edge_table(etab)
            g.remove_edges(rem_edges)
            g.remove_vertices(rem_verts)
            if check_isolated_vertices: g.remove_isolated_vertices()
            if not quiet: print('. ', end='')
            #print('. ', end='', flush=True)
            new_matches = True
            if stats is not None: stats.count_rewrites(name, len(m))
    if not quiet and i>0: print(' {!s} iterations'.format(i))
    if auto_simplify_parallel_edges:
        g.set_auto_simplify(auto_simp_value)
    return i

def pivot_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[ET],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps performing a pivoting rewrite until the list of matches supplied by :func:`match_pivot_parallel` is empty."""
    return simp(g, 'pivot_simp', match_pivot_parallel, pivot, 
                auto_simplify_parallel_edges=True, matchf=matchf, quiet=quiet, stats=stats)

def pivot_gadget_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[ET],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps performing a pivoting rewrite until the list of matches supplied by :func:`match_pivot_gadget` is empty."""
    return simp(g, 'pivot_gadget_simp', match_pivot_gadget, pivot, 
                auto_simplify_parallel_edges=True, matchf=matchf, quiet=quiet, stats=stats)

def pivot_boundary_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[ET],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps performing  a pivoting rewrite until the list of matches supplied by :func:`match_pivot_boundary` is empty."""
    return simp(g, 'pivot_boundary_simp', match_pivot_boundary, pivot, 
                auto_simplify_parallel_edges=True, matchf=matchf, quiet=quiet, stats=stats)

def lcomp_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps performing a local complementation based rewrite until the list of matches supplied by :func:`match_lcomp_parallel` is empty."""
    return simp(g, 'lcomp_simp', match_lcomp_parallel, lcomp, 
                auto_simplify_parallel_edges=True, matchf=matchf, quiet=quiet, stats=stats)

def bialg_simp(g: BaseGraph[VT,ET], quiet:bool=True, stats: Optional[Stats]=None) -> int:
    """Keeps performing a bialgebra rewrite until the list of matches supplied by :func:`match_bialg(_parallel)` is empty."""
    return simp(g, 'bialg_simp', match_bialg_parallel, bialg, quiet=quiet, stats=stats)

def spider_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps performing spider fusion until the list of matches supplied by :func:`match_spider_parallel` is empty."""
    return simp(g, 'spider_simp', match_spider_parallel, spider, matchf=matchf, quiet=quiet, stats=stats)

def id_simp(g: BaseGraph[VT,ET], matchf:Optional[Callable[[VT],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Removes all non-interacting identity vertices found by :func:`match_ids_parallel`."""
    return simp(g, 'id_simp', match_ids_parallel, remove_ids, matchf=matchf, quiet=quiet, stats=stats)

def gadget_simp(g: BaseGraph[VT,ET], matchf: Optional[Callable[[VT],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Removes all phase gadgets that act on the same set of targets found by :func:`match_phase_gadgets`."""
    return simp(g, 'gadget_simp', match_phase_gadgets, merge_phase_gadgets, 
                auto_simplify_parallel_edges=True, matchf=matchf, quiet=quiet, stats=stats)

def supplementarity_simp(g: BaseGraph[VT,ET], quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps performing supplementarity to remove non-Clifford spiders that act on the same set of targets until the list of matches found by :func:`match_supplementarity` is empty."""
    return simp(g, 'supplementarity_simp', match_supplementarity, apply_supplementarity, 
                auto_simplify_parallel_edges=True, quiet=quiet, stats=stats)

def copy_simp(g: BaseGraph[VT,ET], quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Copies 1-ary spiders with 0/pi phase through neighbors.
    WARNING: only use on maximally fused diagrams consisting solely of Z-spiders."""
    return simp(g, 'copy_simp', match_copy, apply_copy, quiet=quiet, stats=stats)

def phase_free_simp(g: BaseGraph[VT,ET], quiet:bool=True, stats:Optional[Stats]=None) -> int:
    '''Performs the following set of simplifications on the graph:
    spider -> bialg'''
    i1 = spider_simp(g, quiet=quiet, stats=stats)
    i2 = bialg_simp(g, quiet=quiet, stats=stats)
    return i1+i2

def basic_simp(g: BaseGraph[VT,ET], matchf: Optional[Callable[[Union[VT, ET]],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps doing the simplifications ``id_simp`` and ``spider_simp`` until none of them can be applied anymore. If
    starting from a circuit, the result should still have causal flow."""
    spider_simp(g, matchf=matchf, quiet=quiet, stats=stats)
    to_gh(g)
    i = 0
    while True:
        i1 = id_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        i2 = spider_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        if i1+i2==0: break
        i += 1
    return i

def interior_clifford_simp(g: BaseGraph[VT,ET], matchf: Optional[Callable[[Union[VT, ET]],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps doing the simplifications ``id_simp``, ``spider_simp``,
    ``pivot_simp`` and ``lcomp_simp`` until none of them can be applied anymore."""
    spider_simp(g, matchf=matchf, quiet=quiet, stats=stats)
    to_gh(g)
    i = 0
    while True:
        i1 = id_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        i2 = spider_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        i3 = pivot_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        i4 = lcomp_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        if i1+i2+i3+i4==0: break
        i += 1
    return i

def clifford_simp(g: BaseGraph[VT,ET], matchf: Optional[Callable[[Union[VT, ET]],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Keeps doing rounds of :func:`interior_clifford_simp` and
    :func:`pivot_boundary_simp` until they can't be applied anymore."""
    i = 0
    while True:
        i += interior_clifford_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        i2 = pivot_boundary_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        if i2 == 0:
            break
    return i

def reduce_scalar(g: BaseGraph[VT,ET], quiet:bool=True, stats:Optional[Stats]=None) -> int:
    """Modification of ``full_reduce`` that is tailered for scalar ZX-diagrams.
    It skips the boundary pivots."""
    i = 0
    while True:
        i1 = id_simp(g, quiet=quiet, stats=stats)
        i2 = spider_simp(g, quiet=quiet, stats=stats)
        i3 = pivot_simp(g, quiet=quiet, stats=stats)
        i4 = lcomp_simp(g, quiet=quiet, stats=stats)
        if i1+i2+i3+i4:
            i += 1
            continue
        i5 = pivot_gadget_simp(g,quiet=quiet, stats=stats)
        i6 = gadget_simp(g, quiet=quiet, stats=stats)
        i7 = copy_simp(g, quiet=quiet, stats=stats)
        if i5 + i6 + i7:
            i += 1
            continue
        i8 = supplementarity_simp(g,quiet=quiet, stats=stats)
        if not i8: break
        i += 1
    return i


def full_reduce(g: BaseGraph[VT,ET], matchf: Optional[Callable[[Union[VT, ET]],bool]]=None, quiet:bool=True, stats:Optional[Stats]=None, paramSafe: Optional[bool]=False) -> None:
    """The main simplification routine of PyZX. It uses a combination of :func:`clifford_simp` and
    the gadgetization strategies :func:`pivot_gadget_simp` and :func:`gadget_simp`. It also attempts to run :func:`supplementarity_simp` and :func:`copy_simp`."""
    if any(g.types()[h] == VertexType.H_BOX for h in g.vertices()):
        raise ValueError("Input graph is not a ZX-diagram as it contains an H-box. "
                         "Maybe call pyzx.hsimplify.from_hypergraph_form(g) first?")
    interior_clifford_simp(g, matchf=matchf, quiet=quiet, stats=stats)
    pivot_gadget_simp(g, matchf=matchf, quiet=quiet, stats=stats)
    while True:
        clifford_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        i, k, l = 0, 0, 0
        if (not paramSafe): i = gadget_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        interior_clifford_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        if (not paramSafe):
            k = copy_simp(g, quiet=quiet, stats=stats)
            l = supplementarity_simp(g,quiet=True, stats=stats)
        j = pivot_gadget_simp(g, matchf=matchf, quiet=quiet, stats=stats)
        if i+j+k+l == 0:
            g.remove_isolated_vertices()
            break

def teleport_reduce(g: BaseGraph[VT,ET], quiet:bool=True, stats:Optional[Stats]=None) -> BaseGraph[VT,ET]:
    """This simplification procedure runs :func:`full_reduce` in a way
    that does not change the graph structure of the resulting diagram.
    The only thing that is different in the output graph are the location and value of the phases."""
    s = Simplifier(g)
    s.full_reduce(quiet=quiet, stats=stats)
    return s.mastergraph


class Simplifier(Generic[VT, ET]):
    """Class used for :func:`teleport_reduce`."""
    def __init__(self, g: BaseGraph[VT,ET]) -> None:
        g.track_phases = True
        self.mastergraph = g.copy()
        self.simplifygraph = g.copy()
        self.simplifygraph.set_phase_master(self)
        self.phantom_phases: Dict[VT, Tuple[VT,int]] = dict()

    def fuse_phases(self,i1:int, i2: int) -> None:
        try:
            v1 = self.mastergraph.vertex_from_phase_index(i1)
            v2 = self.mastergraph.vertex_from_phase_index(i2)
        except ValueError: return
        #self.mastergraph.phase_index[v2] = i1
        p1 = self.mastergraph.phase(v1)
        p2 = self.mastergraph.phase(v2)
        m1 = self.simplifygraph.phase_mult[i1]
        m2 = self.simplifygraph.phase_mult[i2]
        if (p2 == 0 or p2.denominator <= 2): # Deleted vertex contains Clifford phase
            if v2 in self.phantom_phases:
                v3,i3 = self.phantom_phases[v2]
                m2 = cast(Literal[1, -1], m2*self.simplifygraph.phase_mult[i3])
                v2,i2 = v3,i3
                p2 = self.mastergraph.phase(v2)
            else: return
        if (p1 == 0 or p1.denominator <= 2): # Need to save non-Clifford location
            self.simplifygraph.phase_mult[i1] = 1
            if v1 in self.phantom_phases: # Already fused with non-Clifford before
                v3,i3 = self.phantom_phases[v1]
                self.mastergraph.phase_index[v3] = i1
                del self.mastergraph.phase_index[v1]
                p1 = self.mastergraph.phase(v3)
                if (p1+p2).denominator <= 2:
                    del self.phantom_phases[v1]
                v1,i1 = v3,i3
                m1 = cast(Literal[1, -1], m1*self.simplifygraph.phase_mult[i3])
            else:
                self.phantom_phases[v1] = (v2,i2)
                self.simplifygraph.phase_mult[i2] = m2
                return
        if p1.denominator <= 2 or p2.denominator <= 2: raise Exception("Clifford phases here??")
        # Both have non-Clifford phase
        if m1*m2 == 1: phase = (p1 + p2)%2
        else: phase = p1 - p2
        self.mastergraph.set_phase(v1,phase)
        self.mastergraph.set_phase(v2,0)

        self.simplifygraph.phase_mult[i2] = 1

    def full_reduce(self, quiet:bool=True, stats:Optional[Stats]=None) -> None:
        full_reduce(self.simplifygraph,quiet=quiet, stats=stats)



def to_gh(g: BaseGraph[VT,ET],quiet:bool=True) -> None:
    """Turns every red node into a green node by applying a Hadamard to the edges incident to red nodes"""
    ty = g.types()
    for v in g.vertices():
        if ty[v] == VertexType.X:
            g.set_type(v, VertexType.Z)
            for e in g.incident_edges(v):
                et = g.edge_type(e)
                g.set_edge_type(e, toggle_edge(et))


def max_cut(g: BaseGraph[VT,ET], vs0: Optional[Set[VT]]=None, vs1: Optional[Set[VT]]=None) -> Tuple[Set[VT],Set[VT]]:
    """Approximate the MAX-CUT of a graph, starting with an initial partition

    This uses the quadratic-time SG3 heuristic explained by Wang et al in https://arxiv.org/abs/2312.10895 .
    """
    if vs0 is None: vs0 = set()
    if vs1 is None: vs1 = set()
    # print(f'vs0={vs0} vs1={vs1}')
    remaining = set(g.vertices()) - vs0 - vs1
    while len(remaining) > 0:
        score_max = -1
        v_max: Optional[VT] = None
        in0 = True
        for v in remaining:
            wt0 = sum(len(list(g.edges(v,w))) for w in vs1)
            wt1 = sum(len(list(g.edges(v,w))) for w in vs0)
            score = abs(wt0 - wt1)
            if score > score_max:
                # print(f'{v}: score={score}, wt0={wt0}, wt1={wt1}')
                score_max = score
                v_max = v
                in0 = wt0 >= wt1
        # print(f'choosing {v_max} for set {"vs0" if in0 else "vs1"}')
        
        if v_max is None: raise RuntimeError("No max found")
        remaining.remove(v_max)
        if in0: vs0.add(v_max)
        else: vs1.add(v_max)
    return(vs0, vs1)

def to_rg(g: BaseGraph[VT,ET], init_z: Optional[Set[VT]]=None, init_x: Optional[Set[VT]]=None) -> None:
    """Try to eliminate H-edges by turning green nodes red

    This implements a quadratic-time max-cut heuristic to eliminate H-edges. In the future, we may want
    to implement a linear-time version that does a worse job for very large graphs.

    :param g: A ZX-graph.
    :param init_z: An optional set of vertices to make Z.
    :param init_x: An optional set of vertices to make X.
    """

    ty = g.types()

    if init_z is None: init_z = set()
    if init_x is None: init_x = set()

    # initialise boundaries to help max_cut prioritise removing H-edges on the boundary
    for b in g.inputs() + g.outputs():
        v = next(iter(g.neighbors(b)))
        e = g.edge(b, v)
        if g.edge_type(e) == EdgeType.HADAMARD:
            if g.type(v) == VertexType.Z:
                init_z.add(b)
            elif g.type(v) == VertexType.X:
                init_x.add(b)
        else:
            if g.type(v) == VertexType.Z:
                init_x.add(b)
            elif g.type(v) == VertexType.X:
                init_z.add(b)

    vs0, vs1 = max_cut(g, init_z, init_x)
    for v in vs0:
        if ty[v] == VertexType.X:
            g.set_type(v, VertexType.Z)
            for e in g.incident_edges(v):
                g.set_edge_type(e, toggle_edge(g.edge_type(e)))
    for v in vs1:
        if ty[v] == VertexType.Z:
            g.set_type(v, VertexType.X)
            for e in g.incident_edges(v):
                g.set_edge_type(e, toggle_edge(g.edge_type(e)))

def gadgetize(g: BaseGraph[VT,ET], graphlike:bool=True):
    """Convert every non-Clifford phase to a phase gadget"""
    for v in list(g.vertices()):
        p = g.phase(v)
        if not phase_is_clifford(p) and g.vertex_degree(v) > 1:
            y = g.add_vertex(VertexType.Z, -2, g.row(v))
            if graphlike:
                x = g.add_vertex(VertexType.Z, -1, g.row(v))
                g.add_edge((x, y), EdgeType.HADAMARD)
                g.add_edge((v, x), EdgeType.HADAMARD)
            else:
                x = g.add_vertex(VertexType.X, -1, g.row(v))
                g.add_edge((x, y), EdgeType.SIMPLE)
                g.add_edge((v, x), EdgeType.SIMPLE)
            g.set_phase(y, p)
            g.set_phase(v, 0)

def tcount(g: Union[BaseGraph[VT,ET], Circuit]) -> int:
    """Returns the amount of nodes in g that have a non-Clifford phase."""
    if isinstance(g, Circuit):
        return g.tcount()
    count = 0
    phases = g.phases()
    for v in g.vertices():
        if not phase_is_clifford(phases[v]):
            count += 1
    return count


def u3_count(g: BaseGraph[VT, ET]) -> int:
    """Returns the amount of nodes in g that have a non-Clifford phase that is not a multiple of pi/4."""
    count = 0
    phases = g.phases()
    for v in g.vertices():
        if phases[v] != 0 and phases[v].denominator not in (1, 2, 4):
            count += 1
    return count

#The functions below haven't been updated in a while. Use at your own risk.

def simp_iter(
        g: BaseGraph[VT,ET],
        name: str,
        match: Callable[..., List[MatchObject]],
        rewrite: Callable[[BaseGraph[VT,ET],List[MatchObject]],RewriteOutputType[VT,ET]]
        ) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Version of :func:`simp` that instead of performing all rewrites at once, returns an iterator."""
    i = 0
    new_matches = True
    while new_matches:
        i += 1
        new_matches = False
        m = match(g)
        if len(m) > 0:
            etab, rem_verts, rem_edges, check_isolated_vertices = rewrite(g, m)
            g.add_edge_table(etab)
            g.remove_edges(rem_edges)
            g.remove_vertices(rem_verts)
            if check_isolated_vertices: g.remove_isolated_vertices()
            yield g, name+str(i)
            new_matches = True

def pivot_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for the pivot rule, using a list of matches supplied by :func:`match_pivot_parallel`."""
    return simp_iter(g, 'pivot', match_pivot_parallel, pivot)

def pivot_gadget_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for the pivoting rewrite, using a list of matches supplied by :func:`match_pivot_gadget`."""
    return simp_iter(g, 'pivot_gadget', match_pivot_gadget, pivot)

def pivot_boundary_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for the pivoting rewrite, using a list of matches supplied by :func:`match_pivot_boundary`."""
    return simp_iter(g, 'pivot_boundary', match_pivot_boundary, pivot)

def lcomp_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for the local complementation based rewrite rule, using a list of matches supplied by :func:`match_lcomp_parallel`."""
    return simp_iter(g, 'lcomp', match_lcomp_parallel, lcomp)

def bialg_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for the bialgebra rewrite rule, using a list of matches supplied by :func:`match_bialg(_parallel)`."""
    return simp_iter(g, 'bialg', match_bialg_parallel, bialg)

def spider_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for spider fusion, using a list of matches supplied by :func:`match_spider_parallel`."""
    return simp_iter(g, 'spider', match_spider_parallel, spider)

def id_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for removing non-interacting identity vertices, as found by :func:`match_ids_parallel`."""
    return simp_iter(g, 'id', match_ids_parallel, remove_ids)

def gadget_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Returns an iterator for removing phase gadgets that act on the same set of targets, as found by :func:`match_phase_gadgets`"""
    return simp_iter(g, 'gadget', match_phase_gadgets, merge_phase_gadgets)


def clifford_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Like :func:`clifford_simp`, but returns an iterator instead of performing the rewrites all at once."""
    ok = True
    while ok:
        ok = False
        for g, step in interior_clifford_iter(g):
            yield g, step
        for g, step in pivot_boundary_iter(g):
            ok = True
            yield g, step

def interior_clifford_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Like :func:`interior_clifford_simp`, but returns an iterator instead of performing the rewrites all at once."""
    yield from spider_iter(g)
    to_gh(g)
    yield g, "to_gh"
    ok = True
    while ok:
        ok = False
        for g, step in id_iter(g):
            ok = True
            yield g, step
        for g, step in spider_iter(g):
            ok = True
            yield g, step
        for g, step in pivot_iter(g):
            ok = True
            yield g, step
        for g, step in lcomp_iter(g):
            ok = True
            yield g, step

def full_reduce_iter(g: BaseGraph[VT,ET]) -> Iterator[Tuple[BaseGraph[VT,ET],str]]:
    """Like :func:`full_reduce`, but returns an iterator instead of performing the rewrites all at once."""
    yield from interior_clifford_iter(g)
    yield from pivot_gadget_iter(g)
    ok = True
    while ok:
        ok = False
        for g, step in clifford_iter(g):
            yield g, f"clifford -> {step}"
        for g, step in gadget_iter(g):
            ok = True
            yield g, f"gadget -> {step}"
        for g, step in interior_clifford_iter(g):
            yield g, f"interior_clifford -> {step}"
        for g, step in pivot_gadget_iter(g):
            ok = True
            yield g, f"pivot_gadget -> {step}"

def is_graph_like(g: BaseGraph[VT,ET], strict:bool=False) -> bool:
    """Checks if a ZX-diagram is graph-like: 
    only contains Z-spiders which are connected by Hadamard edges.
    If `strict` is True, then also checks that each boundary vertex is connected to a Z-spider,
    and that each Z-spider is connected to at most one boundary."""

    # checks that all spiders are Z-spiders
    for v in g.vertices():
        if g.type(v) not in [VertexType.Z, VertexType.BOUNDARY]:
            return False

    for v1, v2 in itertools.combinations(g.vertices(), 2):
        if not g.connected(v1, v2):
            continue

        # Z-spiders are only connected via Hadamard edges
        if g.type(v1) == VertexType.Z and g.type(v2) == VertexType.Z \
           and g.edge_type(g.edge(v1, v2)) != EdgeType.HADAMARD:
            return False

        # FIXME: no parallel edges

    # no self-loops
    for v in g.vertices():
        if g.connected(v, v):
            return False

    if strict:
        # every I/O is connected to a Z-spider
        bs = [v for v in g.vertices() if g.type(v) == VertexType.BOUNDARY]
        for b in bs:
            if g.vertex_degree(b) != 1 or g.type(list(g.neighbors(b))[0]) != VertexType.Z:
                return False

        # every Z-spider is connected to at most one I/O
        zs = [v for v in g.vertices() if g.type(v) == VertexType.Z]
        for z in zs:
            b_neighbors = [n for n in g.neighbors(z) if g.type(n) == VertexType.BOUNDARY]
            if len(b_neighbors) > 1:
                return False

    return True


def to_graph_like(g: BaseGraph[VT,ET]) -> None:
    """Puts a ZX-diagram in graph-like form."""

    # turn all red spiders into green spiders
    to_gh(g)

    # simplify: remove excess HAD's, fuse along non-HAD edges, remove parallel edges and self-loops
    spider_simp(g, quiet=True)

    # ensure all I/O are connected to a Z-spider
    bs = [v for v in g.vertices() if g.type(v) == VertexType.BOUNDARY]
    for v in bs:

        # if it's already connected to a Z-spider, continue on
        if any([g.type(n) == VertexType.Z for n in g.neighbors(v)]):
            continue

        # have to connect the (boundary) vertex to a Z-spider
        ns = list(g.neighbors(v))
        for n in ns:
            # every neighbor is another boundary or an H-Box
            assert(g.type(n) in [VertexType.BOUNDARY, VertexType.H_BOX])
            if g.type(n) == VertexType.BOUNDARY:
                z1 = g.add_vertex(ty=VertexType.Z)
                z2 = g.add_vertex(ty=VertexType.Z)
                z3 = g.add_vertex(ty=VertexType.Z)
                g.remove_edge(g.edge(v, n))
                g.add_edge((v, z1), edgetype=EdgeType.SIMPLE)
                g.add_edge((z1, z2), edgetype=EdgeType.HADAMARD)
                g.add_edge((z2, z3), edgetype=EdgeType.HADAMARD)
                g.add_edge((z3, n), edgetype=EdgeType.SIMPLE)
            else: # g.type(n) == VertexType.H_BOX
                z = g.add_vertex(ty=VertexType.Z)
                g.remove_edge(g.edge(v, n))
                g.add_edge((v, z), edgetype=EdgeType.SIMPLE)
                g.add_edge((z, n), edgetype=EdgeType.SIMPLE)

    # each Z-spider can only be connected to at most 1 I/O
    vs = list(g.vertices())
    for v in vs:
        if not g.type(v) == VertexType.Z:
            continue
        boundary_ns = [n for n in g.neighbors(v) if g.type(n) == VertexType.BOUNDARY]
        if len(boundary_ns) <= 1:
            continue

        # add dummy spiders for all but one
        for b in boundary_ns[:-1]:
            e = g.edge(v,b)
            if g.edge_type(e) == EdgeType.SIMPLE:
                z1 = g.add_vertex(ty=VertexType.Z,row=0.3*g.row(v)+0.7*g.row(b),qubit=0.3*g.qubit(v)+0.7*g.qubit(b))
                z2 = g.add_vertex(ty=VertexType.Z,row=0.7*g.row(v)+0.3*g.row(b),qubit=0.7*g.qubit(v)+0.3*g.qubit(b))

                g.remove_edge(e)
                g.add_edge((z1, z2), edgetype=EdgeType.HADAMARD)
                g.add_edge((b, z1), edgetype=EdgeType.SIMPLE)
                g.add_edge((z2, v), edgetype=EdgeType.HADAMARD)
            elif g.edge_type(e) == EdgeType.HADAMARD:
                z = g.add_vertex(ty=VertexType.Z,row=0.5*g.row(v)+0.5*g.row(b),qubit=0.5*g.qubit(v)+0.5*g.qubit(b))
                g.remove_edge(e)
                g.add_edge((b,z),EdgeType.SIMPLE)
                g.add_edge((z,v),EdgeType.HADAMARD)

    assert(is_graph_like(g,strict=True))

def unfuse_non_cliffords(g: BaseGraph[VT,ET]) -> None:
    """Unfuses any non-Clifford spider into a magic state connected to a phase-free spider

    Replaces any spider with a non-Clifford phase p and degree n > 1 with a degree 1 spider with phase p connected to
    a degree n+1 spider with phase 0.
    """
    for v in list(g.vertices()):
        ty = g.type(v)
        p = g.phase(v)
        if vertex_is_zx(ty) and not phase_is_clifford(p) and g.vertex_degree(v) > 1:
            v1 = g.add_vertex(ty, qubit=-1, row=g.row(v), phase=p)
            g.set_phase(v, 0)
            g.add_edge((v, v1))

def to_clifford_normal_form_graph(g: BaseGraph[VT,ET]) -> None:
    """Converts a graph that is Clifford into the form described by the right-hand side of eq. (11) of
    *Graph-theoretic Simplification of Quantum Circuits with the ZX-calculus* (https://arxiv.org/abs/1902.03178).
    That is, writes it as a series of layers: 
    Hadamards, phase gates, CZ gates, parity form of Z-spiders to X-spiders, Hadamards, CZ gates, phase gates, Hadamards.
    Changes the graph in place.
    """
    full_reduce(g)
    g.normalize()
    # At this point the only vertices g should have are those directly connected to an input or an output (and not both).
    if any([((g.phase(v)*4) % 2 != 0) for v in g.vertices()]):  # If any phase is not a multiple of 1/2, then this will fail.
        raise ValueError("Specified graph is not Clifford.")

    inputs = list(g.inputs())
    outputs = list(g.outputs())
    v_inputs = [list(g.neighbors(i))[0] for i in inputs] # input vertices should have a unique spider neighbor
    v_outputs = [list(g.neighbors(o))[0] for o in outputs] # input vertices should have a unique spider neighbor
    # create more spacing
    for v in v_inputs:
        g.set_row(v, 3)
    for v in v_outputs:
        g.set_row(v,  5)
    for o in outputs:
        g.set_row(o, 8)
    
    # Separate out the Hadamards 
    for q in range(len(inputs)):
        v = v_inputs[q]
        i = inputs[q]
        e = g.edge(v,i)
        e_type = g.edge_type(e)
        if e_type == EdgeType.HADAMARD or g.phase(v) != 0:
            h = g.add_vertex(VertexType.Z, q, row=1, phase=g.phase(v))
            g.add_edge((i,h),e_type)
            g.add_edge((h,v),EdgeType.SIMPLE)
            g.remove_edge(e)
            g.set_phase(v,0)
            inputs[q] = h
    
    for q in range(len(outputs)):
        v = v_outputs[q]
        o = outputs[q]
        e = g.edge(v,o)
        e_type = g.edge_type(e)
        if e_type == EdgeType.HADAMARD or g.phase(v) != 0:
            h = g.add_vertex(VertexType.Z, q, row=7, phase=g.phase(v))
            g.add_edge((h,o),e_type)
            g.add_edge((v,h),EdgeType.SIMPLE)
            g.remove_edge(e)
            g.set_phase(v,0)
            outputs[q] = h
    
    # Unfuse the czs on the inputs
    czs = []
    cz_qubits = set()
    for q1 in range(len(inputs)):
        for q2 in range(q1+1,len(inputs)):
            if g.connected(v_inputs[q1],v_inputs[q2]):
                g.remove_edge(g.edge(v_inputs[q1],v_inputs[q2]))
                czs.append((q1,q2))
                cz_qubits.add(q1)
                cz_qubits.add(q2)
    cz_v = {}
    for q in cz_qubits:
        w = g.add_vertex(VertexType.Z,q,row=2)
        g.remove_edge(g.edge(inputs[q],v_inputs[q]))
        g.add_edge((inputs[q],w))
        g.add_edge((w,v_inputs[q]))
        cz_v[q] = w
    for q1,q2 in czs:
        g.add_edge((cz_v[q1],cz_v[q2]),EdgeType.HADAMARD)
    
    # Unfuse the czs on the outputs
    czs = []
    cz_qubits = set(range(len(outputs))) # We actually definitely need to add another spider at every position, as we are going to introduce Hadamards everywhere later
    for q1 in range(len(outputs)):
        for q2 in range(q1+1,len(outputs)):
            if g.connected(v_outputs[q1],v_outputs[q2]):
                g.remove_edge(g.edge(v_outputs[q1],v_outputs[q2]))
                czs.append((q1,q2))
                cz_qubits.add(q1)
                cz_qubits.add(q2)
    cz_v = {}
    for q in cz_qubits:
        w = g.add_vertex(VertexType.Z,q,row=6)
        g.remove_edge(g.edge(v_outputs[q],outputs[q]))
        g.add_edge((w,outputs[q]))
        g.add_edge((v_outputs[q],w))
        cz_v[q] = w
    for q1,q2 in czs:
        g.add_edge((cz_v[q1],cz_v[q2]),EdgeType.HADAMARD)
    
    # TODO: re-introduce correct to_rg behaviour here
    #to_rg(g,select=lambda v: v in v_outputs)



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyquil import Program, get_qc
from pyquil.gates import CNOT, S, T, RZ, H, CZ
from pyquil.quil import Pragma
from pyquil.api import LocalQVMCompiler
from pyquil.parser import parse as parse_quil
from pyquil.quilbase import Pragma, Gate
from numpy import pi

from pyzx_param.routing.parity_maps import CNOT_tracker
from pyzx_param.circuit import Circuit

class PyQuilCircuit(CNOT_tracker):

    def __init__(self, architecture, **kwargs):
        """
        Class to represent a PyQuil program to run on/be compiled for the given architecture
        Currently, it assumes the architecture given by create_9x9_square_architecture()

        :param architecture: The Architecture object to adhere to
        """

        super().__init__(**kwargs)
        self.qc = get_qc('9q-square-qvm')
        device = architecture.to_quil_device()
        compiler = LocalQVMCompiler(endpoint=self.qc.compiler.endpoint, device=device)
        self.qc.device = device
        self.qc.compiler = compiler
        self.n_qubits = architecture.n_qubits
        self.program = Program()
        super().__init__(self.n_qubits)
        self.retries = 0
        self.max_retries = 5
        self.compiled_program = None

    def row_add(self, q0, q1):
        """
        Adds a CNOT gate between the given qubit indices q0 and q1
        :param q0: 
        :param q1: 
        """
        self.program += CNOT(q0, q1)
        super().row_add(q0, q1)

    def col_add(self, q0, q1):
        # TODO prepend the CNOT!
        self.program += CNOT(q1, q0)
        super().col_add(q0, q1)

    def count_cnots(self):
        if self.compiled_program is None:
            return super().count_cnots()
        else:
            return self.compiled_cnot_count()

    def compiled_cnot_count(self):
        if self.compiled_program is None:
            self.compile()
        return len([g for g in self.compiled_program if isinstance(g, Gate) and g.name == "CZ"])

    def to_qasm(self):
        if self.compiled_program is None:
            return super().to_qasm()
        circuit = Circuit(self.n_qubits)
        comments = []
        for g in self.compiled_program:
            if isinstance(g, Pragma):
                wiring = " ".join(["//", g.command, "["+g.freeform_string[2:-1]+"]"])
                comments.append(wiring)
            elif isinstance(g, Gate):
                if g.name == "CZ":
                    circuit.add_gate("CZ", g.qubits[0].index, g.qubits[1].index)
                elif g.name == "RX":
                    circuit.add_gate("XPhase", g.qubits[0].index, g.params[0])
                elif g.name == "RZ":
                    circuit.add_gate("ZPhase", g.qubits[0].index, g.params[0])
                else:
                    print("Unsupported gate found!", g)

        qasm = circuit.to_qasm()
        return '\n'.join(comments+[qasm])


    def update_program(self):
        self.program = Program()
        for gate in self.gates:
            if hasattr(gate, "name"):
                if gate.name == "CNOT":
                    self.program += CNOT(gate.control, gate.target)
                elif gate.name == "CZ":
                    self.program += CZ(gate.control, gate.target)
                elif gate.name == "HAD":
                    self.program += H(gate.target)
                elif gate.name == "S":
                    self.program += S(gate.target)
                elif gate.name == "T":
                    self.program += T(gate.target)
                elif gate.name == "T*":
                    self.program += RZ(3*pi/4, gate.target)
                else:
                    print(f"Warning: PyquilCircuit does not currently support the gate '{gate}'.")

    @staticmethod
    def from_CNOT_tracker(circuit, architecture):
        new_circuit = PyQuilCircuit(architecture, n_qubits=circuit.qubits, name=circuit.name)
        new_circuit.gates = circuit.gates
        new_circuit.update_matrix()
        new_circuit.update_program()
        return new_circuit

    @staticmethod
    def from_circuit(circuit, architecture):
        new_circuit = PyQuilCircuit(architecture, n_qubits=circuit.qubits, name=circuit.name)
        new_circuit.gates = circuit.gates
        new_circuit.update_program()
        return new_circuit

    def compile(self):
        """
        Compiles the circuit/program for created quantum computer
        :return: A string that describes the compiled program in quil
        """
        try:
            ep = self.qc.compile(self.program)
            self.retries = 0
            self.compiled_program = parse_quil(ep.program)
            return ep.program
        except KeyError as e:
            print('Oops, retrying to compile.', self.retries)
            if self.retries < self.max_retries:
                self.retries += 1
                return self.compile()
            else:
                raise e
